'''Object to convolve lens PSFs with.'''

from .conf import config
from .convolution import Convolvable
from .coordinates import cart_to_polar

from prysm import mathops as m


class Slit(Convolvable):
    '''Representation of a slit or pair of slits.

    Attributes
    ----------
    orientation : `str`
        orientation of the slit
    width_x : `float`
        x-width of the slit
    width_y : `float`
        y-width of the slit
    '''

    def __init__(self, width, orientation='Vertical', sample_spacing=0.075, samples=0):
        '''Create a new Slit instance.

        Parameters
        ----------
        width : `float`
            the width of the slit in microns
        orientation : `string`, {'Horizontal', 'Vertical', 'Crossed', 'Both'}
            the orientation of the slit; Crossed and Both produce the same results
        sample_spacing : `float`
            spacing of samples in the synthetic image
        samples : `int`
            number of samples per dimension in the synthetic image

        Notes
        -----
        Default of 0 samples allows quick creation for convolutions without
        generating the image; use samples > 0 for an actual image.

        '''
        w = width / 2

        if samples > 0:
            ext = samples / 2 * sample_spacing
            x, y = m.linspace(-ext, ext, samples), m.linspace(-ext, ext, samples)
            arr = m.zeros((samples, samples))
        else:
            arr, x, y = None, m.zeros(2), m.zeros(2)

        # paint in the slit
        if orientation.lower() in ('v', 'vert', 'vertical'):
            if samples > 0:
                arr[:, abs(x) < w] = 1
            self.orientation = 'Vertical'
            self.width_x = width
            self.width_y = 0
        elif orientation.lower() in ('h', 'horiz', 'horizontal'):
            if samples > 0:
                arr[abs(y) < w, :] = 1
            self.width_x = 0
            self.width_y = width
            self.orientation = 'Horizontal'
        elif orientation.lower() in ('b', 'both', 'c', 'crossed'):
            if samples > 0:
                arr[abs(y) < w, :] = 1
                arr[:, abs(x) < w] = 1
            self.orientation = 'Crossed'
            self.width_x, self.width_y = width, width

        super().__init__(arr, x, y, has_analytic_ft=True)

    def analytic_ft(self, unit_x, unit_y):
        '''Analytic fourier transform of a slit.

        Parameters
        ----------
        unit_x : `numpy.ndarray`
            sample points in x frequency axis
        unit_y : `numpy.ndarray`
            sample points in y frequency axis

        Returns
        -------
        `numpy.ndarray`
            2D numpy array containing the analytic fourier transform

        '''
        xq, yq = m.meshgrid(unit_x, unit_y)
        if self.width_x > 0 and self.width_y > 0:
            return (m.sinc(xq * self.width_x) +
                    m.sinc(yq * self.width_y)).astype(config.precision)
        elif self.width_x > 0 and self.width_y is 0:
            return m.sinc(xq * self.width_x).astype(config.precision)
        else:
            return m.sinc(yq * self.width_y).astype(config.precision)


class Pinhole(Convolvable):
    '''Representation of a pinhole.

    Attributes
    ----------
    width : `float`
        diameter of the pinhole

    '''
    def __init__(self, width, sample_spacing=0.025, samples=0):
        '''Produce a Pinhole.

        Parameters
        ----------
        width : `float`
            the width of the pinhole
        sample_spacing : `float`
            spacing of samples in the synthetic image
        samples : `int`
            number of samples per dimension in the synthetic image

        Notes
        -----
        Default of 0 samples allows quick creation for convolutions without
        generating the image; use samples > 0 for an actual image.

        '''
        self.width = width

        # produce coordinate arrays
        if samples > 0:
            ext = samples / 2 * sample_spacing
            x, y = m.linspace(-ext, ext, samples), m.linspace(-ext, ext, samples)
            xv, yv = m.meshgrid(x, y)
            w = width / 2
            # paint a circle on a black background
            arr = m.zeros((samples, samples))
            arr[m.sqrt(xv**2 + yv**2) < w] = 1
        else:
            arr, x, y = None, m.zeros(2), m.zeros(2)

        super().__init__(data=arr, unit_x=x, unit_y=y, has_analytic_ft=True)

    def analytic_ft(self, unit_x, unit_y):
        '''Analytic fourier transform of a slit.

        Parameters
        ----------
        unit_x : `numpy.ndarray`
            sample points in x frequency axis
        unit_y : `numpy.ndarray`
            sample points in y frequency axis

        Returns
        -------
        `numpy.ndarray`
            2D numpy array containing the analytic fourier transform

        '''
        xq, yq = m.meshgrid(unit_x, unit_y)

        # factor of pi corrects for jinc being modulo pi
        # factor of 2 converts radius to diameter
        rho = m.sqrt(xq**2 + yq**2) * self.width * 2 * m.pi
        return m.jinc(rho).astype(config.precision)


class SiemensStar(Convolvable):
    '''Representation of a Siemen's star object.

    Attributes
    ----------
    spokes : `int`
        number of spokes present in the star

    '''
    def __init__(self, spokes, sinusoidal=True, background='black', sample_spacing=2, samples=256):
        '''Produces a Siemen's Star.

        Parameters
        ----------
        spokes : `int`
            number of spokes in the star.
        sinusoidal : `bool`
            if True, generates a sinusoidal Siemen' star, else, generates a bar/block siemen's star
        background : 'string', {'black', 'white'}
            background color
        sample_spacing : `float`
            spacing of samples, in microns
        samples : `int`
            number of samples per dimension in the synthetic image

        Raises
        ------
        ValueError
            background other than black or white

        '''
        relative_width = 0.9
        self.spokes = spokes

        # generate a coordinate grid
        x = m.linspace(-1, 1, samples)
        y = m.linspace(-1, 1, samples)
        xx, yy = m.meshgrid(x, y)
        rv, pv = cart_to_polar(xx, yy)
        ext = sample_spacing * samples / 2
        ux, uy = m.arange(-ext, ext, sample_spacing), m.arange(-ext, ext, sample_spacing)

        # generate the siemen's star as a (rho,phi) polynomial
        arr = m.cos(spokes / 2 * pv)

        if not sinusoidal:  # make binary
            arr[arr < 0] = -1
            arr[arr > 0] = 1

        # scale to (0,1) and clip into a disk
        arr = (arr + 1) / 2
        if background.lower() in ('b', 'black'):
            arr[rv > relative_width] = 0
        elif background.lower() in ('w', 'white'):
            arr[rv > relative_width] = 1
        else:
            raise ValueError('invalid background color')

        super().__init__(data=arr, unit_x=ux, unit_y=uy, has_analytic_ft=False)


class TiltedSquare(Convolvable):
    '''Represents a tilted square for e.g. slanted-edge MTF calculation.'''

    def __init__(self, angle=8, background='white', sample_spacing=2, samples=256):
        '''Create a new TitledSquare instance.

        Parameters
        ----------
        angle : `float`
            angle in degrees to tilt w.r.t. the x axis
        background : `string`
            white or black; the square will be the opposite color of the background
        sample_spacing : `float`
            spacing of samples
        samples : `int`
            number of samples

        '''
        radius = 0.3
        if background.lower() == 'white':
            arr = m.ones((samples, samples))
            fill_with = 0
        else:
            arr = m.zeros((samples, samples))
            fill_with = 1

        # TODO: optimize by working with index numbers directly and avoid
        # creation of X,Y arrays for performance.
        x = m.linspace(-0.5, 0.5, samples)
        y = m.linspace(-0.5, 0.5, samples)
        xx, yy = m.meshgrid(x, y)
        sf = samples * sample_spacing

        # TODO: convert inline operation to use of rotation matrix
        angle = m.radians(angle)
        xp = xx * m.cos(angle) - yy * m.sin(angle)
        yp = xx * m.sin(angle) + yy * m.cos(angle)
        mask = (abs(xp) < radius) * (abs(yp) < radius)
        arr[mask] = fill_with
        super().__init__(data=arr, unit_x=x * sf, unit_y=y * sf, has_analytic_ft=False)


class SlantedEdge(Convolvable):
    """Representation of a slanted edge."""

    def __init__(self, angle=8, contrast=0.9, crossed=False, sample_spacing=2, samples=256):
        '''Create a new TitledSquare instance.

        Parameters
        ----------
        angle : `float`
            angle in degrees to tilt w.r.t. the y axis
        contrast : `float`
            difference between minimum and maximum values in the image
        crossed : `bool`, optional
            whether to make a single edge (crossed=False) or pair of crossed edges (crossed=True)
            aka a "BMW target"
        sample_spacing : `float`
            spacing of samples
        samples : `int`
            number of samples

        '''
        diff = (1 - contrast) / 2
        arr = m.full((samples, samples), 1 - diff)
        x = m.linspace(-0.5, 0.5, samples)
        y = m.linspace(-0.5, 0.5, samples)
        xx, yy = m.meshgrid(x, y)
        sf = samples * sample_spacing

        angle = m.radians(angle)
        xp = xx * m.cos(angle) - yy * m.sin(angle)
        # yp = xx * m.sin(angle) + yy * m.cos(angle)  # do not need this
        mask = xp > 0  # single edge
        if crossed:
            mask = xp > 0  # set of 4 edges
            upperright = mask & m.rot90(mask)
            lowerleft = m.rot90(upperright, 2)
            mask = upperright | lowerleft

        arr[mask] = diff
        self.contrast = contrast
        self.black = diff
        self.white = 1 - diff
        super().__init__(data=arr, unit_x=x * sf, unit_y=y * sf, has_analytic_ft=False)



'''A repository of fringe Zernike aberration descriptions used to model pupils of optical systems.'''
from collections import defaultdict

from .conf import config
from .pupil import Pupil
from .coordinates import make_rho_phi_grid, cart_to_polar
from .util import rms, share_fig_ax, sort_xy

from prysm import mathops as m

from prysm import _zernike as z


zernmap = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 8,
    9: 9,
    10: 10,
    11: 11,
    12: 12,
    13: 13,
    14: 14,
    15: 15,
    16: 16,
    17: 17,
    18: 18,
    19: 19,
    20: 20,
    21: 21,
    22: 22,
    23: 23,
    24: 24,
    25: 25,
    26: 26,
    27: 27,
    28: 28,
    29: 29,
    30: 30,
    31: 31,
    32: 32,
    33: 33,
    34: 34,
    35: 35,
    36: 36,
    37: 37,
    38: 38,
    39: 39,
    40: 40,
    41: 41,
    42: 42,
    43: 43,
    44: 44,
    45: 45,
    46: 46,
    47: 47,
    48: 48,
 }


def fzname(idx, base=1):
    """Return the name of a Fringe Zernike with the given index and base."""
    return z.zernikes[zernmap[idx-base]].name


def fzset_to_magnitude_angle(coefs):
    """Convert Fringe Zernike polynomial set to a magnitude and phase representation."""

    def mkary():  # default for defaultdict
        return m.zeros(2)

    # make a list of names to go with the coefficients
    names = [fzname(i, base=0) for i in range(len(coefs))]
    combinations = defaultdict(mkary)

    # for each name and coefficient, make a len 2 array.  Put the Y or 0 degree values in the first slot
    for coef, name in zip(coefs, names):
        if name.endswith(('X', 'Y', '°')):
            newname = ' '.join(name.split(' ')[:-1])
            if name.endswith('Y'):
                combinations[newname][0] = coef
            elif name.endswith('X'):
                combinations[newname][1] = coef
            elif name[-2] == '5':  # 45 degree case
                combinations[newname][1] = coef
            else:
                combinations[newname][0] = coef
        else:
            combinations[name][0] = coef

    # print(combinations)
    # now go over the combinations and compute the L2 norms and angles
    for name in combinations:
        ovals = combinations[name]
        magnitude = m.sqrt((ovals**2).sum())
        phase = m.degrees(m.arctan2(*ovals))
        values = (magnitude, phase)
        combinations[name] = values

    return dict(combinations)  # cast to regular dict for return


class FZCache(object):
    def __init__(self):
        self.normed = defaultdict(dict)
        self.regular = defaultdict(dict)

    def get_zernike(self, number, norm, samples):
        if norm is True:
            target = self.normed
        else:
            target = self.regular

        try:
            zern = target[samples][number]
        except KeyError:
            rho, phi = make_rho_phi_grid(samples, aligned='y')
            func = z.zernikes[zernmap[number]]
            zern = func(rho, phi)
            if norm is True:
                zern *= func.norm

            target[samples][number] = zern.copy()

        return zern

    def __call__(self, number, norm, samples):
        return self.get_zernike(number, norm, samples)

    def clear(self, *args):
        self.normed = defaultdict(dict)
        self.regular = defaultdict(dict)


class FringeZernike(Pupil):
    def __init__(self, *args, **kwargs):
        if args is not None:
            if len(args) is 0:
                self.coefs = m.zeros(len(zernmap), dtype=config.precision)
            else:
                self.coefs = m.asarray([*args[0]], dtype=config.precision)

        self.normalize = False
        pass_args = {}

        self.base = config.zernike_base
        try:
            bb = kwargs['base']
            if bb > 1:
                raise ValueError('It violates convention to use a base greater than 1.')
            elif bb < 0:
                raise ValueError('It is nonsensical to use a negative base.')
            self.base = bb
        except KeyError:
            # user did not specify base
            pass

        if kwargs is not None:
            for key, value in kwargs.items():
                if key[0].lower() == 'z':
                    idx = int(key[1:])  # strip 'Z' from index
                    self.coefs[idx - self.base] = value
                elif key in ('norm'):
                    self.normalize = True
                elif key.lower() == 'base':
                    self.base = value
                else:
                    pass_args[key] = value

        super().__init__(**pass_args)

    def build(self):
        '''Uses the wavefront coefficients stored in this class instance to
            build a wavefront model.

        Returns
        -------
        self.phase : `numpy.ndarray`
            arrays containing the phase associated with the pupil
        self.fcn : `numpy.ndarray`
            array containing the wavefunction of the pupil plane

        '''
        # build a coordinate system over which to evaluate this function
        self.phase = m.zeros((self.samples, self.samples), dtype=config.precision)
        for term, coef in enumerate(self.coefs):
            # short circuit for speed
            if coef == 0:
                continue
            self.phase += coef * zcache(term, self.normalize, self.samples)

        return self

    @property
    def magnitudes(self):
        """Returns the magnitude and angles of the zernike components in this wavefront."""
        return fzset_to_magnitude_angle(self.coefs)

    def barplot(self, orientation='h', buffer=1, zorder=3, fig=None, ax=None):
        """Creates a barplot of coefficients and their names.

        Parameters
        ----------
        orientation : `str`, {'h', 'v', 'horizontal', 'vertical'}
            orientation of the plot
        buffer : `float`, optional
            buffer to use around the left and right (or top and bottom) bars
        zorder : `int`, optional
            zorder of the bars.  Use zorder > 3 to put bars in front of gridlines
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import pyplot as plt
        fig, ax = share_fig_ax(fig, ax)

        coefs = m.asarray(self.coefs)
        idxs = m.asarray(range(len(coefs))) + self.base
        names = [fzname(i) for i in (idxs - self.base)]
        lab = f'{self.zaxis_label} [{self.phase_unit}]'
        lims = (idxs[0] - buffer, idxs[-1] + buffer)
        if orientation.lower() in ('h', 'horizontal'):
            vmin, vmax = coefs.min(), coefs.max()
            drange = vmax - vmin
            offset = drange * 0.01

            ax.bar(idxs, self.coefs, zorder=zorder)
            plt.xticks(idxs, names, rotation=90)
            for i in idxs:
                ax.text(i, offset, str(i), ha='center')
            ax.set(ylabel=lab, xlim=lims)
        else:
            ax.barh(idxs, self.coefs, zorder=zorder)
            plt.yticks(idxs, names)
            for i in idxs:
                ax.text(0, i, str(i), ha='center')
            ax.set(xlabel=lab, ylim=lims)
        return fig, ax

    def barplot_magnitudes(self, orientation='h', sort=False, buffer=1, zorder=3, fig=None, ax=None):
        """Create a barplot of magnitudes of coefficient pairs and their names.

        E.g., astigmatism will get one bar.

        Parameters
        ----------
        orientation : `str`, {'h', 'v', 'horizontal', 'vertical'}
            orientation of the plot
        sort : `bool`, optional
            whether to sort the zernikes in descending order
        buffer : `float`, optional
            buffer to use around the left and right (or top and bottom) bars
        zorder : `int`, optional
            zorder of the bars.  Use zorder > 3 to put bars in front of gridlines
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import pyplot as plt

        magang = fzset_to_magnitude_angle(self.coefs)
        mags = [m[0] for m in magang.values()]
        names = magang.keys()
        idxs = list(range(len(names)))

        if sort:
            mags, names = sort_xy(mags, names)
            mags = list(reversed(mags))
            names = list(reversed(names))
        lab = f'{self.zaxis_label} [{self.phase_unit}]'
        lims = (idxs[0] - buffer, idxs[-1] + buffer)
        fig, ax = share_fig_ax(fig, ax)
        if orientation.lower() in ('h', 'horizontal'):
            ax.bar(idxs, mags, zorder=zorder)
            plt.xticks(idxs, names, rotation=90)
            ax.set(ylabel=lab, xlim=lims)
        else:
            ax.barh(idxs, mags, zorder=zorder)
            plt.yticks(idxs, names)
            ax.set(xlabel=lab, ylim=lims)
        return fig, ax

    def top_n(self, n=5):
        """Identify the top n terms in the wavefront.

        Parameters
        ----------
        n : `int`, optional
            identify the top n terms.

        Returns
        -------
        `list`
            list of tuples (magnitude, index, term)

        """
        coefs = m.asarray(self.coefs)
        coefs_work = abs(coefs)
        oidxs = m.arange(len(coefs)) + self.base  # "original indexes"
        idxs = m.argpartition(coefs_work, -n)[-n:]  # argpartition does some magic to identify the top n (unsorted)
        idxs = idxs[m.argsort(coefs_work[idxs])[::-1]]  # use argsort to sort them in ascending order and reverse
        big_terms = coefs[idxs]  # finally, take the values from the
        big_idxs = oidxs[idxs]
        names = [fzname(i) for i in idxs]
        return list(zip(big_terms, big_idxs, names))

    def barplot_topn(self, n=5, orientation='h', buffer=1, zorder=3, fig=None, ax=None):
        """Plot the top n terms in the wavefront.

        Parameters
        ----------
        n : `int`, optional
            plot the top n terms.
        orientation : `str`, {'h', 'v', 'horizontal', 'vertical'}
            orientation of the plot
        buffer : `float`, optional
            buffer to use around the left and right (or top and bottom) bars
        zorder : `int`, optional
            zorder of the bars.  Use zorder > 3 to put bars in front of gridlines
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import pyplot as plt

        topn = self.top_n(n)
        magnitudes = [n[0] for n in topn]
        names = [n[2] for n in topn]
        idxs = range(len(names))

        fig, ax = share_fig_ax(fig, ax)

        lab = f'{self.zaxis_label} [{self.phase_unit}]'
        lims = (idxs[0] - buffer, idxs[-1] + buffer)
        if orientation.lower() in ('h', 'horizontal'):
            ax.bar(idxs, magnitudes, zorder=zorder)
            plt.xticks(idxs, names, rotation=90)
            ax.set(ylabel=lab, xlim=lims)
        else:
            ax.barh(idxs, magnitudes, zorder=zorder)
            plt.yticks(idxs, names)
            ax.set(xlabel=lab, ylim=lims)
        return fig, ax

    def truncate(self, n):
        """Truncate the wavefront to the first n terms.

        Parameters
        ----------
        n : `int`
            number of terms to keep.

        Returns
        -------
        `self`
            modified FringeZernike instance.

        """
        if n > len(self.coefs):
            return self
        else:
            self.coefs = self.coefs[:n]
            self.build()
            self.mask(self._mask, self.mask_target)
            return self

    def truncate_topn(self, n):
        """Truncate the pupil to only the top n terms.

        Parameters
        ----------
        n : `int`
            number of parameters to keep

        Returns
        -------
        `self`
            modified FringeZernike instance.

        """
        topn = self.top_n(n)
        new_coefs = m.zeros(len(self.coefs), dtype=config.precision)
        for coef in topn:
            mag, index, *_ = coef
            new_coefs[index-self.base] = mag

        self.coefs = new_coefs
        self.build()
        self.mask(self._mask, self.mask_target)
        return self

    def __repr__(self):
        '''Pretty-print pupil description.'''
        if self.normalize is True:
            header = 'rms normalized Fringe Zernike description with:\n\t'
        else:
            header = 'Fringe Zernike description with:\n\t'

        strs = []
        for number, (coef, func) in enumerate(zip(self.coefs, z.zernikes)):
            # skip 0 terms
            if coef == 0:
                continue

            # positive coefficient, prepend with +
            if m.sign(coef) == 1:
                _ = '+' + f'{coef:.3f}'
            # negative, sign comes from the value
            else:
                _ = f'{coef:.3f}'

            # create the name
            name = f'Z{number+self.base} - {func.name}'

            strs.append(' '.join([_, name]))
        body = '\n\t'.join(strs)

        footer = f'\n\t{self.pv:.3f} PV, {self.rms:.3f} RMS'
        return f'{header}{body}{footer}'


def fit(data, x=None, y=None, rho=None, phi=None, terms=16, norm=False, residual=False, round_at=6):
    '''Fits a number of Zernike coefficients to provided data by minimizing
        the root sum square between each coefficient and the given data.  The
        data should be uniformly sampled in an x,y grid.

    Parameters
    ----------
    data : `numpy.ndarray`
        data to fit to.

    x : `numpy.ndarray`, optional
        x coordinates, same shape as data
    y : `numpy.ndarray`, optional
        y coordinates, same shape as data
    rho : `numpy.ndarray`, optional
        radial coordinates, same shape as data
    phi : `numpy.ndarray`, optional
        azimuthal
    terms : `int`, optional
        number of terms to fit, fits terms 0~terms
    norm : `bool`, optional
        if True, normalize coefficients to unit RMS value
    residual : `bool`, optional
        if True, return a tuple of (coefficients, residual)
    round_at : `int`
        decimal place to round values at.

    Returns
    -------
    coefficients : `numpy.ndarray`
        an array of coefficients matching the input data.
    residual : `float`
        RMS error between the input data and the fit.

    Raises
    ------
    ValueError
        too many terms requested.

    '''
    if terms > len(zernmap):
        raise ValueError(f'number of terms must be less than {len(zernmap)}')

    data = data.T  # transpose to mimic transpose of zernikes

    # precompute the valid indexes in the original data
    pts = m.isfinite(data)

    if x is None and rho is None:
        # set up an x/y rho/phi grid to evaluate Zernikes on
        rho, phi = make_rho_phi_grid(*reversed(data.shape))
        rho = rho[pts].flatten()
        phi = phi[pts].flatten()
    elif rho is None:
        rho, phi = cart_to_polar(x, y)
        rho, phi = rho[pts].flatten(), phi[pts].flatten()

    # compute each Zernike term
    zernikes = []
    for i in range(terms):
        func = z.zernikes[zernmap[i]]
        base_zern = func(rho, phi)
        if norm:
            base_zern *= func.norm
        zernikes.append(base_zern)
    zerns = m.asarray(zernikes).T

    # use least squares to compute the coefficients
    meas_pts = data[pts].flatten()
    coefs = m.lstsq(zerns, meas_pts, rcond=None)[0]
    if round_at is not None:
        coefs = coefs.round(round_at)

    if residual is True:
        components = []
        for zern, coef in zip(zernikes, coefs):
            components.append(coef * zern)

        _fit = m.asarray(components)
        _fit = _fit.sum(axis=0)
        rmserr = rms(data[pts].flatten() - _fit)
        return coefs, rmserr
    else:
        return coefs


zcache = FZCache()
config.chbackend_observers.append(zcache.clear)



"""A repository of seidel aberration descriptions used to model pupils of optical systems.
"""
from functools import lru_cache

from .conf import config
from .pupil import Pupil

from prysm import mathops as m


class Seidel(Pupil):
    """Seidel pupil description.

    Attributes
    ----------
    coefs : `list`
        list of coefficient values
    eqns : `list`
        list of Wxxx expressions
    field : `int`
        field point associated with this pupil
    phase : `numpy.ndarray`
        phase errors of this pupil
    fcn : `numpy.ndarray`
        wavefunction of this pupil

    """

    def __init__(self, *args, **kwargs):
        """Create a new Seidel instance.

        Parameters
        ----------
        **kwargs
            - 'Wxxx' any argument of this form corresponding to an H H Hopkins wavefront polynomial
            - {'field', 'relative_field', 'h'}, field height

        """
        self.eqns = []
        self.coefs = []
        pass_args = {}
        self.field = 1
        if kwargs is not None:
            for key, value in kwargs.items():
                if key[0].lower() == 'w' and len(key) == 4:
                    self.eqns.append(wexpr_to_opd_expr(key))
                    self.coefs.append(value)
                elif key.lower() in ('field', 'relative_field', 'h'):
                    self.field = value
                else:
                    pass_args[key] = value

        super().__init__(**pass_args)

    def build(self):
        """Use the wavefront coefficients stored in this class instance to build a wavefront model.

        Returns
        -------
        self.phase : `numpy.ndarray`
            phase errors over (x,y) coordinates
        self.fcn : `numpy.ndarray`
            wavefunction over (x,y) coordinates

        """
        mathexprs = ['m.zeros((self.samples, self.samples))']
        for term, coef in zip(self.eqns, self.coefs):
            mathexprs.append(str(coef) + '*(' + term + ')')

        # pull the field point into the namespace our expression wants
        self._gengrid()
        H = self.field
        rho, phi = self.rho, self.phi

        # compute the pupil phase and wave function
        self.phase = eval('+'.join(mathexprs)).astype(config.precision)
        return self.phase

    def __repr__(self):
        """Describe object.

        Returns
        -------
        `str`
            object description

        """
        return str(self.__dict__)


@lru_cache()
def wexpr_to_opd_expr(Wxxx):
    """Convert a W notation to a string with numpy code to evaluate for pupil phase.

    Parameters
    ----------
    Wxxx : `string`
        A string of the form "W000," "W131", etc.

    Returns
    -------
    `string`
        Contains typed numpy expressions to be evaluated to return phase

    """
    # pop the W off and separate the characters
    _ = list(Wxxx[1:])
    H, rho, phi = _[0], _[1], _[2]
    # .format converts to bytecode, f-strings do not.  Micro-optimization here
    return 'H**{0} * rho**{1} * m.cos(phi)**{2}'.format(H, rho, phi)



''' A repository of standard zernike aberration descriptions used to model pupils of
optical systems.
'''

from .conf import config
from .pupil import Pupil

from prysm import mathops as m

_names = (
    'Z0  - Piston / Bias',
    'Z1  - Tilt X',
    'Z2  - Tilt Y',
    'Z3  - Primary Astigmatism 00deg',
    'Z4  - Defocus / Power',
    'Z5  - Primary Astigmatism 45deg',
    'Z6  - Primary Trefoil X',
    'Z7  - Primary Coma X',
    'Z8  - Primary Coma Y',
    'Z9  - Primary Trefoil Y',
    'Z10 - Primary Tetrafoil X',
    'Z11 - Secondary Astigmatism 00deg',
    'Z12 - Primary Spherical',
    'Z13 - Secondary Astigmatism 45deg',
    'Z14 - Primary Tetrafoil Y',
    'Z15 - Primary Pentafoil X',
    'Z16 - Secondary Trefoil X',
    'Z17 - Secondary Coma X',
    'Z18 - Secondary Coma Y',
    'Z19 - Secondary Trefoil Y',
    'Z20 - Primary Pentafoil Y',
    'Z21 - Primary Hexafoil X',
    'Z22 - Secondary Tetrafoil X',
    'Z23 - Tertiary Astigmatism 00deg',
    'Z24 - Secondary Spherical',
    'Z25 - Tertariary Astigmatism 45deg',
    'Z26 - Secondary Tetrafoil Y',
    'Z27 - Primary Hexafoil Y',
    'Z28 - Primary Heptafoil X',
    'Z29 - Secondary Pentafoil X',
    'Z30 - Tertiary Trefoil X',
    'Z31 - Tertiary Coma X',
    'Z32 - Tertiary Coma Y',
    'Z33 - Tertiary Trefoil Y',
    'Z34 - Secondary Pentafoil Y',
    'Z35 - Primary Heptafoil Y',
    'Z36 - Primary Octafoil X',
    'Z37 - Secondary Hexafoil X',
    'Z38 - Tertiary Tetrafoil X',
    'Z39 - Quarternary Astigmatism 00deg',
    'Z40 - Tertiary Spherical',
    'Z41 - Quarternary Astigmatism 45deg',
    'Z42 - Tertiary Tetrafoil Y',
    'Z43 - Secondary Hexafoil Y',
    'Z44 - Primary Octafoil Y',
    'Z45 - Primary Nonafoil X',
    'Z46 - Secondary Heptafoil X',
    'Z47 - Tertiary Pentafoil X',
)


@m.jit
def Z0(rho, phi):
    return m.ones(rho.shape)


@m.vectorize
def Z1(rho, phi):
    return rho * m.cos(phi)


@m.vectorize
def Z2(rho, phi):
    return rho * m.sin(phi)


@m.vectorize
def Z3(rho, phi):
    return rho**2 * m.cos(2 * phi)


@m.vectorize
def Z4(rho, phi):
    return 2 * rho ** 2 - 1


@m.vectorize
def Z5(rho, phi):
    return rho**2 * m.sin(2 * phi)


@m.vectorize
def Z6(rho, phi):
    return rho**3 * m.cos(3 * phi)


@m.vectorize
def Z7(rho, phi):
    return (3 * rho**3 - 2 * rho) * m.cos(phi)


@m.vectorize
def Z8(rho, phi):
    return (3 * rho**3 - 2 * rho) * m.sin(phi)


@m.vectorize
def Z9(rho, phi):
    return rho**3 * m.sin(3 * phi)


@m.vectorize
def Z10(rho, phi):
    return rho**4 * m.cos(4 * phi)


@m.vectorize
def Z11(rho, phi):
    return (4 * rho**4 - 3 * rho**2) * m.cos(2 * phi)


@m.vectorize
def Z12(rho, phi):
    return -6 * rho**2 + 6 * rho**4 + 1


@m.vectorize
def Z13(rho, phi):
    return (4 * rho**4 - 3 * rho**2) * m.sin(2 * phi)


@m.vectorize
def Z14(rho, phi):
    return rho**4 * m.sin(4 * phi)


@m.vectorize
def Z15(rho, phi):
    return rho**5 * m.cos(5 * phi)


@m.vectorize
def Z16(rho, phi):
    return (5 * rho**5 - 4 * rho**3) * m.cos(3 * phi)


@m.vectorize
def Z17(rho, phi):
    return (10 * rho**5 - 12 * rho**3 + 3 * rho) * m.cos(phi)


@m.vectorize
def Z18(rho, phi):
    return (10 * rho**5 - 12 * rho**3 + 3 * rho) * m.sin(phi)


@m.vectorize
def Z19(rho, phi):
    return (5 * rho**5 - 4 * rho**3) * m.sin(3 * phi)


@m.vectorize
def Z20(rho, phi):
    return rho**5 * m.cos(5 * phi)


@m.vectorize
def Z21(rho, phi):
    return rho**6 * m.cos(6 * phi)


@m.vectorize
def Z22(rho, phi):
    return (6 * rho**6 - 5 * rho**4) * m.cos(4 * phi)


@m.vectorize
def Z23(rho, phi):
    return (15 * rho**6 - 20 * rho**4 + 6 * rho**2) * m.cos(2 * phi)


@m.vectorize
def Z24(rho, phi):
    return 20 * rho**6 - 30 * rho**4 + 12 * rho**2 - 1


@m.vectorize
def Z25(rho, phi):
    return (15 * rho**6 - 20 * rho**4 + 6 * rho**2) * m.sin(2 * phi)


@m.vectorize
def Z26(rho, phi):
    return (6 * rho**6 - 5 * rho**4) * m.sin(4 * phi)


@m.vectorize
def Z27(rho, phi):
    return rho**6 * m.sin(6 * phi)


@m.vectorize
def Z28(rho, phi):
    return rho**6 * m.cos(7 * phi)


@m.vectorize
def Z29(rho, phi):
    return (7 * rho**7 - 6 * rho**5) * m.cos(5 * phi)


@m.vectorize
def Z30(rho, phi):
    return (21 * rho**7 - 30 * rho**5 + 10 * rho**3) * m.cos(3 * phi)


@m.vectorize
def Z31(rho, phi):
    return (35 * rho**7 - 60 * rho**5 + 30 * rho**3 - 4 * rho) * m.cos(phi)


@m.vectorize
def Z32(rho, phi):
    return (35 * rho**7 - 60 * rho**5 + 30 * rho**3 - 4 * rho) * m.sin(phi)


@m.vectorize
def Z33(rho, phi):
    return (21 * rho**7 - 30 * rho**5 + 10 * rho**3) * m.sin(3 * phi)


@m.vectorize
def Z34(rho, phi):
    return (7 * rho**7 - 6 * rho**5) * m.sin(5 * phi)


@m.vectorize
def Z35(rho, phi):
    return rho**7 * m.sin(7 * phi)


@m.vectorize
def Z36(rho, phi):
    return rho**8 * m.cos(8 * phi)


@m.vectorize
def Z37(rho, phi):
    return (8 * rho**8 - 7 * rho**6) * m.cos(6 * phi)


@m.vectorize
def Z38(rho, phi):
    return (28 * rho**8 - 42 * rho**6 + 15 * rho**4) * m.cos(4 * phi)


@m.vectorize
def Z39(rho, phi):
    return (56 * rho**8 - 105 * rho**6 + 60 * rho**4 - 10 * rho**2) * m.cos(2 * phi)


@m.vectorize
def Z40(rho, phi):
    return 70 * rho**8 - 140 * rho**7 + 90 * rho**4 - 20 * rho**2 + 1


@m.vectorize
def Z41(rho, phi):
    return (56 * rho**8 - 105 * rho**6 + 60 * rho**4 - 10 * rho**2) * m.cos(2 * phi)


@m.vectorize
def Z42(rho, phi):
    return (28 * rho**8 - 42 * rho**6 + 15 * rho**4) * m.sin(4 * phi)


@m.vectorize
def Z43(rho, phi):
    return (8 * rho**8 - 7 * rho**6) * m.sin(6 * phi)


@m.vectorize
def Z44(rho, phi):
    return rho**8 * m.sin(8 * phi)


@m.vectorize
def Z45(rho, phi):
    return rho**9 * m.cos(9 * phi)


@m.vectorize
def Z46(rho, phi):
    return (9 * rho**9 - 8 * rho**7) * m.cos(7 * phi)


@m.vectorize
def Z47(rho, phi):
    return (36 * rho**9 - 56 * rho**7 + 21 * rho**5) * m.cos(5 * phi)


zernfcns = {
    0.0: Z0,
    1.0: Z1,
    2.0: Z2,
    3.0: Z3,
    4.0: Z4,
    5.0: Z5,
    6.0: Z6,
    7.0: Z7,
    8.0: Z8,
    9.0: Z9,
    10.0: Z10,
    11.0: Z11,
    12.0: Z12,
    13.0: Z13,
    14.0: Z14,
    15.0: Z15,
    16.0: Z16,
    17.0: Z17,
    18.0: Z18,
    19.0: Z19,
    20.0: Z20,
    21.0: Z21,
    22.0: Z22,
    23.0: Z23,
    24.0: Z24,
    25.0: Z25,
    26.0: Z26,
    27.0: Z27,
    28.0: Z28,
    29.0: Z29,
    30.0: Z30,
    31.0: Z31,
    32.0: Z32,
    33.0: Z33,
    34.0: Z34,
    35.0: Z35,
    36.0: Z36,
    37.0: Z37,
    38.0: Z38,
    39.0: Z39,
    40.0: Z40,
    41.0: Z41,
    42.0: Z42,
    43.0: Z43,
    44.0: Z44,
    45.0: Z45,
    46.0: Z46,
    47.0: Z47,
}


def zernwrapper(term, rho, phi):
    ''' Wraps the Z0..Z48 functions.
    '''
    return zernfcns[term](rho, phi)


class StandardZernike(Pupil):
    '''Standard Zernike pupil description

    Properties:
        Inherited from :class:`Pupil`, please see that class.

    Instance Methods:
        build: computes the phase and wavefunction for the pupil.  This method
            is automatically called by the constructor, and does not regularly
            need to be changed by the user.

    Private Instance Methods:
        none

    Static Methods:
        none

    '''
    def __init__(self, *args, **kwargs):
        ''' Creates a StandardZernike Pupil object.

        Args:
            samples (int): number of samples across pupil diameter.

            wavelength (float): wavelength of light, in um.

            epd: (float): diameter of the pupil, in mm.

            opd_unit (string): unit OPD is expressed in.  One of:
                ($\lambda$, waves, $\mu m$, microns, um, nm , nanometers).

            base (`int`): 0 or 1, adjusts the base index of the polynomial
                expansion.

            Zx (float): xth standard zernike coefficient, in range [0,47], 0-base.

        Returns:
            StandardZernike.  A new :class:`StandardZernike` pupil instance.

        Notes:
            Supports multiple syntaxes:
                - args: pass coefficients as a list, including terms up to the highest given Z-index.
                        e.g. passing [1,2,3] will be interpreted as Z0=1, Z1=2, Z3=3.

                - kwargs: pass a named set of zernike terms.
                          e.g. StandardZernike(Z0=1, Z1=2, Z2=3)

            Supports unit conversion, can pass kwarg:
                - opd_unit='nm': coefficients are expressed in units of nm

            The kwargs syntax overrides the args syntax.

        '''

        if args is not None:
            if len(args) is 0:
                self.coefs = [0] * len(zernfcns)
            else:
                self.coefs = [*args[0]]
        else:
            self.coefs = [0] * len(zernfcns)

        pass_args = {}

        self.base = config.zernike_base
        try:
            bb = kwargs['base']
            if bb > 1:
                raise ValueError('It violates convention to use a base greater than 1.')
            self.base = bb
        except KeyError:
            # user did not specify base
            pass

        if kwargs is not None:
            for key, value in kwargs.items():
                if key[0].lower() == 'z':
                    idx = int(key[1:])  # strip 'Z' from index
                    self.coefs[idx - self.base] = value
                elif key.lower() == 'base':
                    self.base = value
                else:
                    pass_args[key] = value

        super().__init__(**pass_args)

    def build(self):
        # construct an equation for the phase of the pupil
        # build a coordinate system over which to evaluate this function
        self._gengrid()
        self.phase = m.zeros((self.samples, self.samples), dtype=config.precision)
        for term, coef in enumerate(self.coefs):
            # short circuit for speed
            if coef == 0:
                continue
            self.phase += coef * zernwrapper(term=term,
                                             rho=self.rho,
                                             phi=self.phi)

        self._correct_phase_units()
        return self.phase, self.fcn

    def __repr__(self):
        '''Pretty-print pupil description
        '''
        header = 'Standard Zernike description with:\n\t'

        strs = []
        for number, (coef, name) in enumerate(zip(self.coefs, _names)):
            # skip 0 terms
            if coef == 0:
                continue

            # positive coefficient, prepend with +
            if m.sign(coef) == 1:
                _ = '+' + f'{coef:.3f}'
            # negative, sign comes from the value
            else:
                _ = f'{coef:.3f}'

            # adjust term numbers
            if self.base is 1:
                if number > 9:  # two-digit term
                    name_lcl = ''.join([name[0],
                                        str(int(name[1:3]) + 1),
                                        name[3:]])
                else:
                    name_lcl = ''.join([name[0],
                                        str(int(name[1]) + 1),
                                        name[2:]])
            else:
                name_lcl = name

            strs.append(' '.join([_, name_lcl]))
        body = '\n\t'.join(strs)

        footer = f'\n\t{self.pv:.3f} PV, {self.rms:.3f} RMS'
        return f'{header}{body}{footer}'


def fit(data, num_terms=16, rms_norm=False, round_at=6):
    ''' Fits a number of zernike coefficients to provided data by minimizing
        the root sum square between each coefficient and the given data.  The
        data should be uniformly sampled in an x,y grid.

    Args:

        data (`numpy.ndarray`): data to fit to.

        num_terms (`int`): number of terms to fit, fits terms 0~num_terms.

        rms_norm (`bool`): if true, normalize coefficients to unit RMS value.

        round_at (`int`): decimal place to round values at.

    Returns:
        numpy.ndarray: an array of coefficients matching the input data.

    '''
    if num_terms > len(zernfcns):
        raise ValueError(f'number of terms must be less than {len(zernfcns)}')

    # precompute the valid indexes in the original data
    pts = m.isfinite(data)

    # set up an x/y rho/phi grid to evaluate zernikes on
    x, y = m.linspace(-1, 1, data.shape[1]), m.linspace(-1, 1, data.shape[0])
    xv, yv = m.meshgrid(x, y)
    rho = m.sqrt(xv**2 + yv**2)[pts].flatten()
    phi = m.arctan2(xv, yv)[pts].flatten()

    # compute each zernike term
    zernikes = []
    for i in range(num_terms):
        zernikes.append(zernwrapper(i, rms_norm, rho, phi))
    zerns = m.asarray(zernikes).T

    # use least squares to compute the coefficients
    coefs = m.linalg.lstsq(zerns, data[pts].flatten())[0]
    return coefs.round(round_at)



"""A submodule which imports and exports math functions from different libraries.

The intend is to make the backend for prysm interoperable, allowing users to
utilize more high performance engines if they have them installed, or fall
back to more widely available options in the case that they do not.
"""
import numpy as np

from scipy.special import j1, j0  # NOQA

from prysm.conf import config

# numba funcs
try:
    from numba import jit, vectorize
    numba_installed = True
except ImportError:
    # if Numba is not installed, create the jit decorator and have it return the
    # original function.
    def jit(signature_or_function=None, locals={}, target='cpu', cache=False, **options):
        """Passthrough duplicate of numba jit.

        Parameters
        ----------
        signature_or_function : None, optional
            a function or signature to compile
        locals : dict, optional
            local variables for the compiled code
        target : str, optional
            architecture to compile for
        cache : bool, optional
            whether to cache compilation to disk
        **options
            various options

        Returns
        -------
        callable
            a callable

        """
        if signature_or_function is None:
            def _jit(function):
                return function
            return _jit
        else:
            return signature_or_function

    vectorize = jit
    numba_installed = False

# cuda
try:
    import cupy as cp
    from cupy import fuse
    assert cp
    cuda_compatible = True
except ImportError:
    cuda_compatible = False

    def fuse(*args, **kwargs):
        if 'func' in kwargs:
            return kwargs['func']
        else:
            def wrapper(function):
                return function

            return wrapper

allfuncs = set((
    'sort',
    'sqrt',
    'sin',
    'cos',
    'tan',
    'arctan',
    'arctan2',
    'arccos',
    'arcsin',
    'sinc',
    'radians',
    'exp',
    'log',
    'log10',
    'linspace',
    'meshgrid',
    'angle',
    'zeros',
    'ones',
    'empty',
    'sign',
    'isfinite',
    'asarray',
    'arange',
    'stack',
    'mean',
    'unique',
    'flip',
    'swapaxes',
    'rollaxis',
    'concatenate',
    'cumsum',
    'any',
    'all',
    'isfinite',
    'isnan',
    'ceil',
    'floor',
    'outer',
    'inner',
    'argmin',
    'argmax',
    'allclose',
    'frombuffer',
    'count_nonzero',
    'trapz',
    'hanning',
    'full',
    'pad',
    'minimum',
    'maximum',
    'fromstring',
    'rot90',
    'flipud',
    'fliplr',
    'ascontiguousarray',
    'argpartition',
    'argsort',
    'degrees',
))

allfuncs_cupy_missing = set((
    'searchsorted',
    'gradient',
    ))

fftfuncs = frozenset((
    'fft2',
    'ifft2',
    'fftshift',
    'ifftshift',
    'fftfreq',
))

linalgfuncs = frozenset((
    'lstsq',
))

constants = frozenset((
    'ndarray',
    'uint8',
    'uint16',
    'uint32',
    'int32',
    'int64',
    'float32',
    'float64',
    'complex64',
    'complex128',
    'newaxis',
    'pi',
    'nan',
    'inf',
))


def jinc(r):
    """Jinc.

    Parameters
    ----------
    r : `number`
        radial distance

    Returns
    -------
    `float`
        the value of j1(x)/x for x != 0, 0.5 at 0

    """
    if r < 1e-8 and r > -1e-8:  # value of jinc for x < 1/2 machine precision  is 0.5
        return 0.5
    else:
        return j1(r) / r


if numba_installed is True:
    # one day split numba jit and numpy jit
    jinc = np.vectorize(jinc)
else:
    jinc = np.vectorize(jinc)


def change_backend(to):
    if to == 'cu':
        if not cuda_compatible:
            raise ValueError('installation lacks cuda support.')
        else:
            target_base = 'cupy'
            target_fft = 'cupy.fft'
            target_linalg = 'cupy.linalg'
            # target_scipy = 'cupyx.scipy'

    elif to == 'np':
        target_base = 'numpy'
        target_fft = 'numpy.fft'
        target_linalg = 'numpy.linalg'
        # target_scipy = 'scipy'

        # two sets of functionality unavailable via cupy
        for func in allfuncs_cupy_missing:
            exec(f'from {target_base} import {func}')
            globals()[func] = eval(func)

        for func in linalgfuncs:
            exec(f'from {target_linalg} import {func}')
            globals()[func] = eval(func)

    for func in allfuncs:
        exec(f'from {target_base} import {func}')
        globals()[func] = eval(func)

    for const in constants:
        exec(f'from {target_base} import {const}')
        globals()[const] = eval(const)

    for func in fftfuncs:
        exec(f'from {target_fft} import {func}')
        globals()[func] = eval(func)


config.chbackend_observers.append(change_backend)
config.backend = config.backend  # trigger import of math functions



''' Model of optical systems
'''
import warnings
from functools import partial
from copy import deepcopy

from scipy.optimize import minimize

from .conf import config
from .seidel import Seidel
from .psf import PSF
from .otf import MTF
from .util import share_fig_ax
from .thinlens import image_displacement_to_defocus
from .mtf_utils import MTFvFvF

from prysm import mathops as m


class Lens(object):
    ''' Represents a lens or optical system.
    '''
    def __init__(self, **kwargs):
        ''' Create a new Lens object.

        Args:
            efl (`float`): Effective Focal Length.

            fno (`float`): Focal Ratio.

            pupil_magnification (`float`): Ratio of exit pupil to entrance pupil
                diameter.

            aberrations (`dict`): A dictionary

            fields (`iterable`): A set of relative field points to analyze (symmetric)

            fov_x (`float`): half Field of View in X

            fov_y (`float`): half Field of View in Y

            fov_unit (`string`): unit for field of view.  mm, degrees, etc.

            wavelength (`float`): wavelength of light, in um.

            samples (`float`): samples in the pupil plane used to compute wavefronts.
        '''
        efl = 1
        fno = 1
        pupil_magnification = 1
        ab = dict()
        fields = [0, 1]
        fov_x = 0
        fov_y = 21.64
        fov_unit = 'mm'
        wavelength = 0.55
        samples = 128
        if kwargs is not None:
            for key, value in kwargs.items():
                kl = key.lower()
                if kl == 'efl':
                    efl = value
                elif kl == 'fno':
                    fno = value
                elif kl == 'pupil_magnification':
                    pupil_magnification = value
                elif kl in ('aberrations', 'abers', 'abs'):
                    ab = value
                elif kl == 'fields':
                    fields = value
                elif kl == 'fov_x':
                    fov_x = value
                elif kl == 'fov_y':
                    fov_y = value
                elif kl == 'fov_unit':
                    fov_unit = value
                elif kl == 'wavelength':
                    wavelength = value
                elif kl == 'samples':
                    samples = value

        if efl < 0:
            warnings.warn('''
                Negative focal lengths are treated as positive for fresnel
                diffraction propogation to function correctly.  In the context
                of these simulations a positive and negative focal length are
                functionally equivalent and the provide value has had its sign
                flipped.
                ''')
            efl *= -1
        if fno < 0:
            raise ValueError('f/# must by definition be positive')

        self.efl = efl
        self.fno = fno
        self.pupil_magnification = pupil_magnification
        self.epd = efl / fno
        self.xpd = self.epd * pupil_magnification
        self.aberrations = ab
        self.fields = fields
        self.fov_x = fov_x
        self.fov_y = fov_y
        self.fov_unit = fov_unit
        self.wavelength = wavelength
        self.samples = samples

    ####### analytically setting aberrations -----------------------------------

    def autofocus(self, field_index=0):
        ''' Adjusts the W020 aberration coefficient to maximize the MTF at a
            given field index.

        Args:
            field_index (`int`): index of the field to maximize MTF at.

        Returns:
            `Lens` self.
        '''
        coefs = self.aberrations.copy()
        try:
            # try to access the W020 aberration
            float(coefs['W020'])
        except KeyError:
            # if it is not set, make it 0
            coefs['W020'] = 0.0

        def opt_fcn(self, coefs, w020):
            # shift the defocus term appropriately
            abers = coefs.copy()
            abers['W020'] += w020
            pupil = Seidel(**abers, epd=self.epd, samples=self.samples, h=self.fields[field_index])

            # cost value (to be minimized) is RMS wavefront
            return pupil.rms

        opt_fcn = partial(opt_fcn, self, coefs)

        new_defocus = minimize(opt_fcn, x0=0, method='Powell')
        coefs['W020'] += float(new_defocus['x'])
        self.aberrations = coefs.copy()
        return self

    ####### analytically setting aberrations -----------------------------------

    ####### data generation ----------------------------------------------------

    def psf_vs_field(self, num_pts):
        ''' Generates a list of PSFs as a function of field.

        Args:
            num_pts (`int`): number of points to generate a PSF for.

        Returns:
            `list` containing the PSF objects.

        '''
        self._uniformly_spaced_fields(num_pts)
        psfs = []
        for idx in range(num_pts):
            psfs.append(self._make_psf(idx))
        return psfs

    def mtf_vs_field(self, num_pts, freqs=[10, 20, 30, 40, 50]):
        ''' Generates a 2D array of MTF vs field values for the given spatial
            frequencies.

        Args:
            num_pts (`int`): Number of points to compute the MTF at.

            freqs (`iterable`): set of frequencies to compute at.

        Returns:
            `tuple` containing:

                `numpy.ndarray` (Tan) a 3D ndnarray where the columns
                    correspond to fields and the rows correspond to spatial
                    frequencies.

                `numpy.ndarray` (Sag) a 3D ndnarray where the columns
                    correspond to fields and the rows correspond to spatial
                    frequencies.

        '''
        self._uniformly_spaced_fields(num_pts)
        mtfs_t = m.empty((num_pts, len(freqs)))
        mtfs_s = m.empty((num_pts, len(freqs)))
        for idx in range(num_pts):
            mtf = self._make_mtf(idx)
            vals_t = mtf.exact_polar(freqs, 0)
            vals_s = mtf.exact_polar(freqs, 90)
            mtfs_t[idx, :] = vals_t
            mtfs_s[idx, :] = vals_s

        return mtfs_s, mtfs_t

    ####### data generation ----------------------------------------------------

    ####### plotting -----------------------------------------------------------

    def plot_psf_vs_field(self, num_pts, fig=None, axes=None, axlim=25):
        ''' Creates a figure showing the evolution of the PSF over the field
            of view.

        Args:
            num_pts (`int`): Number of points between (0,1) to create a PSF for

        Returns:
            `tuple` containing:

                `matplotlib.pyplot.figure` figure containing the plots.

                `list` the axes the plots are placed in.

        '''
        psfs = self.psf_vs_field(num_pts)
        fig, axes = share_fig_ax(fig, axes, numax=num_pts, sharex=True, sharey=True)

        for idx, (psf, axis) in enumerate(zip(psfs, axes)):
            show_labels = False
            show_colorbar = False
            if idx == 0:
                show_labels = True
            elif idx == num_pts - 1:
                show_colorbar = True
            psf.plot2d(fig=fig, ax=axis, axlim=axlim,
                       show_axlabels=show_labels, show_colorbar=show_colorbar)

        fig_width = 15
        fig.set_size_inches(fig_width, fig_width / num_pts)
        fig.tight_layout()
        return fig, axes

    def plot_mtf_vs_field(self, num_pts, freqs=[10, 20, 30, 40, 50], title='MTF vs Field', minorgrid=True, fig=None, ax=None):
        ''' Generates a plot of the MTF vs Field for the lens.

        Args:
            num_pts (`int`): number of field points to evaluate.

            freqs (`iterable`): frequencies to evaluate the MTF at.

            fig (`matplotlib.pyplot.figure`): figure to plot inside.

            ax (`matplotlib.pyplot.axis`): axis to plot ini.

        Return:
            `tuple` containing:

                `matplotlib.pyplot.figure` figure containing the plot.

                `matplotlib.pyplot.axis` axis containing the plot.

        '''
        data_s, data_t = self.mtf_vs_field(num_pts, freqs)
        flds_abs = m.linspace(0, self.fov_y, num_pts)
        fig, ax = share_fig_ax(fig, ax)
        for i in range(len(freqs)):
            ln, = ax.plot(flds_abs, data_s[:, i], lw=3, ls='--')
            ax.plot(flds_abs, data_t[:, i], lw=3, color=ln.get_color(), label=f'{freqs[i]}lp/mm')

        ax.plot(0, 0, color='k', ls='--', label='Sag')
        ax.plot(0, 0, color='k', label='Tan')
        # todo: respect units of `self`

        if minorgrid is True:
            ax.set_yticks([0.1, 0.3, 0.5, 0.7, 0.9], minor=True)
            ax.grid(True, which='minor')

        ax.set(xlim=(0, self.fov_y), xlabel='Image Height [mm]',
               ylim=(0, 1), ylabel='MTF [Rel. 1.0]',
               title=title)
        ax.legend()

        return fig, ax

    def plot_mtf_thrufocus(self, field_index, focus_range, numpts, freqs, fig=None, ax=None):
        focus, mtfs = self._make_mtf_thrufocus(field_index, focus_range, numpts)
        t = []
        s = []
        for mtf in mtfs:
            t.append(mtf.exact_polar(freqs, 0))
            s.append(mtf.exact_polar(freqs, 90))

        t, s = m.asarray(t), m.asarray(s)
        fig, ax = share_fig_ax(fig, ax)
        for idx, freq in enumerate(freqs):
            l, = ax.plot(focus, t[:, idx], lw=2, label=freq)
            ax.plot(focus, s[:, idx], lw=2, ls='--', c=l.get_color())
        ax.legend(title=r'$\nu$ [cy/mm]')
        ax.set(xlim=(focus[0], focus[-1]), xlabel=r'Defocus [$\mu m$]',
               ylim=(0, 1), ylabel='MTF [Rel. 1.0]',
               title='Through Focus MTF')

        return fig, ax

    ####### plotting -----------------------------------------------------------

    ####### helpers ------------------------------------------------------------

    def _make_pupil(self, field_index):
        ''' Generates the pupil for a given field

        Args:
            field_index (`int`): index of the desired field in the self.fields
                iterable.

        Returns:
            `Pupil` a pupil object.
        '''
        return Seidel(**self.aberrations,
                      epd=self.epd,
                      h=self.fields[field_index],
                      wavelength=self.wavelength,
                      samples=self.samples)

    def _make_psf(self, field_index):
        ''' Generates the psf for a given field

        Args:
            field_index (`int`): index of the desired field in the self.fields
                iterable.

        Returns:
            `PSF` a psf object.
        '''
        p = self._make_pupil(field_index=field_index)
        return PSF.from_pupil(p, self.efl)

    def _make_mtf(self, field_index):
        ''' Generates the mtf for a given field

        Args:
            field_index (`int`): index of the desired field in the self.fields
                iterable.

        Returns:
            `MTF` an MTF object.
        '''
        pp = self._make_psf(field_index=field_index)
        return MTF.from_psf(pp)

    def _make_mtf_thrufocus(self, field_index, focus_range, num_pts):
        ''' Makes a list of MTF objects corresponding to different focus shifts
            for the lens.  Focusrange will be applied symmetrically.

        Args:
            field_index: (`int`): index of the desired field in the self.fields
                iterable.

            focus_range: (`float`): focus range, in microns.

            num_pts (`int`): number of points to compute MTF at.  Note that for
                and even number of points, the zero defocus point will not be
                sampled.

        Returns:
            list of `MTF` objects.

        '''
        # todo: parallelize
        focus_shifts = m.linspace(-focus_range, focus_range, num_pts)
        defocus_wvs = image_displacement_to_defocus(focus_shifts, self.fno, self.wavelength)

        mtfs = []
        pupil = self._make_pupil(field_index)
        for defocus in defocus_wvs:
            defocus_p = Seidel(W020=defocus, epd=self.epd,
                               samples=self.samples, wavelength=self.wavelength)
            psf = PSF.from_pupil(pupil.merge(defocus_p), self.efl)
            mtfs.append(MTF.from_psf(psf))
        return focus_shifts, mtfs

    def _make_mtf_vs_field_vs_focus(self, num_fields, focus_range, num_focus, freqs):
        ''' TODO: docstring
        '''
        self._uniformly_spaced_fields(num_fields)
        net_mtfs = [None] * num_fields
        for idx in range(num_fields):
            focus, net_mtfs[idx] = self._make_mtf_thrufocus(idx, focus_range, num_focus)

        fields = (self.fields[-1] * self.fov_y) * m.linspace(0, 1, num_fields)
        t_cube = m.empty((num_focus, num_fields, len(freqs)))
        s_cube = m.empty((num_focus, num_fields, len(freqs)))
        for idx, mtfs in enumerate(net_mtfs):
            for idx2, submtf in enumerate(mtfs):
                t = submtf.exact_polar(freqs, 0)
                s = submtf.exact_polar(freqs, 90)
                t_cube[idx2, idx, :] = t
                s_cube[idx2, idx, :] = s

        TCube = MTFvFvF(data=t_cube, focus=focus, field=fields, freq=freqs, azimuth='Tan')
        SCube = MTFvFvF(data=s_cube, focus=focus, field=fields, freq=freqs, azimuth='Sag')
        return TCube, SCube

    def _uniformly_spaced_fields(self, num_pts):
        ''' Changes the `fields` property to n evenly spaced points from 0~1.

        Args:
            num_pts (`int`): number of points.

        Returns:
            self.

        '''
        _ = m.arange(0, num_pts, dtype=config.precision)
        flds = _ / _.max()
        self.fields = flds
        return self

    ####### helpers ------------------------------------------------------------

    def clone(self):
        ''' Makes a deep copy of this Lens instance.

        Returns:
            `Lens` a new Lens instance.
        '''
        ret = Lens()
        ret.__dict__ = deepcopy(self.__dict__)
        return ret

    def __repr__(self):
        return (f'Lens with properties:\n\t'
                f'efl: {self.efl}\n\t'
                f'f/#: {self.fno}\n\t'
                f'pupil mag: {self.pupil_magnification}\n\t'
                'Aberrations:\n\t\t'
                f'{str(self.aberrations)}')


def _spherical_defocus_from_monochromatic_mtf(lens, frequencies, mtf_s, mtf_t):
    ''' Uses nonlinear optimization to set the W020, W040, W060, and W080
        coefficients in a lens model based on MTF measurements taken on the
        optical axis.

    Args:
        lens (`Lens`): a lens object.

        frequencies (`iterable`): A set of frequencies the provided MTF values
            correspond to.

        mtf_s (`iterable`): A set of sagittal MTF measurements of equal length
            to the frequencies argument.

        mtf_t (`iterable`): A set of tangential MTF measurements of equal length
            to the frequencies argument.

    Returns:
        `Lens` A new lens object with its aberrations field modified with new
            spherical coefficients.

    '''
    work_lens = lens.clone()

    fcn = partial(_spherical_cost_fcn_raw, frequencies,
                  mtf_s, mtf_t, work_lens)

    results = minimize(fcn, [0, 0, 0, 0], method='Powell')
    W020, W040, W060, W080 = results['x']
    work_lens.aberrations['W020'] = W020
    work_lens.aberrations['W040'] = W040
    work_lens.aberrations['W060'] = W060
    work_lens.aberrations['W080'] = W080
    return work_lens


def _spherical_cost_fcn_raw(frequencies, truth_s, truth_t, lens, abervalues):
    ''' TODO - document.  partial() should be used on this and scipy.minimize'd

        abervalues - array of [W020, W040, W060, W080]
    '''
    pupil = Seidel(epd=lens.epd, samples=lens.samples,
                   W020=abervalues[0],
                   W040=abervalues[1],
                   W060=abervalues[2],
                   W080=abervalues[3])
    psf = PSF.from_pupil(pupil, efl=lens.efl)
    mtf = MTF.from_psf(psf)
    synth_t = mtf.exact_polar(frequencies, 0)
    synth_s = mtf.exact_polar(frequencies, 90)

    truth = m.stack((truth_s, truth_t))
    synth = m.stack((synth_s, synth_t))

    return ((truth - synth) ** 2).sum()



"""File readers for various commercial instruments."""
import os
import re
import struct
import codecs
import datetime
import calendar

from .conf import config
from prysm import mathops as m


def read_file_stream_or_path(path_or_file):
    try:
        with codecs.open(path_or_file, mode='r', encoding='cp1252') as fid:
            data = codecs.encode(fid.read(), 'utf-8').decode('utf-8')
    except (FileNotFoundError, TypeError):  # FNF -- file object, TypeError -- file_like
        try:
            path_or_file.seek(0)
            raw = path_or_file.read()
            data = codecs.encode(raw, 'utf-8').decode('utf-8')
        except TypeError:  # opened in bytes mode
            data = raw.decode('cp1252')
        except AttributeError:
            data = path_or_file  # TODO: avoid duplicate
    except (AttributeError, UnicodeDecodeError):
        data = path_or_file

    return data


def is_mtfvfvf_file(file):
    """Read MTF vs Field vs Focus data from a Trioptics .txt dump.

    Parameters
    ----------
    file : `str` or path_like or file_like
        file to read from, if string of file body, must provide filename

    Returns:
    boolean : `bool`
        if the file is an MTFvFvF file
    data : `str`
        contents of the file
    `
    """
    data = read_file_stream_or_path(file)
    if data.startswith('ImgHeight'):
        return True, data
    else:
        return False, data


def read_trioptics_mtfvfvf(file, filename=None):
    """Read MTF vs Field vs Focus data from a Trioptics .txt dump.

    Parameters
    ----------
    file : `str` or path_like or file_like
        file to read from, if string of file body, must provide filename
    filename : `str`, optional
        name of file; used to select tan/sag if file is given as contents

    Returns
    -------
    `MTFvFvF`
        MTF vs Field vs Focus object

    """
    if filename is None:
        with open(file, 'r') as fid:
            lines = fid.readlines()
    else:
        lines = file.splitlines()
        file = filename

    if str(file)[-7:-4] == 'Tan':
        azimuth = 'Tan'
    else:
        azimuth = 'Sag'

    imghts, objangs, focusposes, mtfs = [], [], [], []
    for meta, data in zip(lines[0::2], lines[1::2]):  # iterate 2 lines at a time
        metavalues = meta.split()
        imght, objang, focuspos, freqpitch = metavalues[1::2]
        mtf_raw = data.split()[1:]  # first element is "MTF"
        mtf = m.asarray(mtf_raw, dtype=config.precision)
        imghts.append(imght)
        objangs.append(objang)
        focusposes.append(focuspos)
        mtfs.append(mtf)

    focuses = m.unique(m.asarray(focusposes, dtype=config.precision))
    focuses = (focuses - m.mean(focuses)) * 1e3
    imghts = m.unique(m.asarray(imghts, dtype=config.precision))
    freqs = m.arange(len(mtfs[0]), dtype=config.precision) * float(freqpitch)
    data = m.swapaxes(m.asarray(mtfs).reshape(len(focuses), len(imghts), len(freqs)), 0, 1)
    return {
        'data': data,
        'focus': focuses,
        'field': imghts,
        'freq': freqs,
        'azimuth': azimuth
    }


def read_trioptics_mtf_vs_field(file, metadata=False):
    """Read tangential and sagittal MTF data from a Trioptics .mht file.

    Parameters
    ----------
    file : `str` or path_like or file_like
        contents of a file, path_like to the file, or file object
    metadata : `bool`
        whether to also extract and return metadata

    Returns
    -------
    `dict`
        dictionary with keys of freq, field, tan, sag

    """
    data = read_file_stream_or_path(file)
    data = data[:len(data)//10]  # only search in a subset of the file for speed

    # compile a pattern that will search for the image heights in the file and extract
    fields_pattern = re.compile(f'MTF=09{os.linesep}(.*?){os.linesep}Legend=09', flags=re.DOTALL)
    fields = fields_pattern.findall(data)[0]  # two copies, only need 1st

    # make a pattern that will search for and extract the tan and sag MTF data.  The match will
    # return two copies; one for vs imght, one for vs angle.  Only keep half the matches.
    tan_pattern = re.compile(r'Tan(.*?)=97', flags=re.DOTALL)
    sag_pattern = re.compile(r'Sag(.*?)=97', flags=re.DOTALL)
    tan, sag = tan_pattern.findall(data), sag_pattern.findall(data)
    endpt = len(tan) // 2
    tan, sag = tan[:endpt], sag[:endpt]

    # now extract the freqs from the tan data
    freqs = m.asarray([float(s.split('(')[0][1:]) for s in tan])

    # lastly, extract the floating point tan and sag data
    # also take fields, to the 4th decimal place (nearest .1um)
    # reformat T/S to 2D arrays with indices of (freq, field)
    tan = m.asarray([s.split('=09')[1:-1] for s in tan], dtype=config.precision)
    sag = m.asarray([s.split('=09')[1:-1] for s in sag], dtype=config.precision)
    fields = m.asarray(fields.split('=09')[0:-1], dtype=config.precision).round(4)
    res = {
        'freq': freqs,
        'field': fields,
        'tan': tan,
        'sag': sag,
    }
    if metadata is True:
        return {**res, **parse_trioptics_metadata(data)}
    else:
        return res


def read_trioptics_mtf(file, metadata=False):
    """Read MTF data from a Trioptics data file.

    Parameters
    ----------
    file : `str` or path_like or file_like
        contents of a file, path_like to the file, or file object
    metadata : `bool`
        whether to also extract and return metadata

    Returns
    -------
    `dict`
        dictionary with keys focus, wavelength, freq, tan, sag
        if metadata=True, also has keys in the return of
        `io.parse_trioptics_metadata`.

    """
    data = read_file_stream_or_path(file)
    data = data[:len(data)//10]

    # compile regex scanners to grab wavelength, focus, and frequency information
    # in addition to the T, S MTF data.
    # lastly, compile a scanner to cut the file after the end of the "MTF Sagittal" scanner
    focus_scanner = re.compile(r'Focus Position  : (\-?\d+\.\d+) mm')
    data_scanner = re.compile(r'\r\n(\d+\.?\d?)=09\r\n(\d+\.\d+)=09')
    sag_scanner = re.compile(r'Measurement Table: MTF vs. Frequency \( Sagittal \)')
    blockend_scanner = re.compile(r'  _____ =20')

    sagpos, cutoff = sag_scanner.search(data).end(), None
    for blockend in blockend_scanner.finditer(data):
        if blockend.end() > sagpos and cutoff is None:
            cutoff = blockend.end()

    # get focus and wavelength
    focus_pos = float(focus_scanner.search(data).group(1))

    # simultaneously grab frequency and MTF
    result = data_scanner.findall(data[:cutoff])
    freqs, mtfs = [], []
    for dat in result:
        freqs.append(float(dat[0]))
        mtfs.append(dat[1])

    breakpt = len(mtfs) // 2
    t = m.asarray(mtfs[:breakpt], dtype=config.precision)
    s = m.asarray(mtfs[breakpt:], dtype=config.precision)
    freqs = tuple(freqs[:breakpt])

    res = {
        'focus': focus_pos,
        'freq': freqs,
        'tan': t,
        'sag': s,
    }
    if metadata is True:
        return {**res, **parse_trioptics_metadata(data)}
    else:
        return res


def parse_trioptics_metadata(file_contents):
    """Read metadata from the contents of a Trioptics .mht file.

    Parameters:
    -----------
    file_contents : `str`
        contents of a .mht file.

    Returns:
    --------
    `dict`
        dictionary with keys:
            - operator
            - time
            - sample_id
            - instrument
            - instrument_sn
            - collimator
            - wavelength
            - efl
            - obj_angle
            - focus_pos
            - azimuth

    """
    data = file_contents[750:1500]  # skip large section to make regex faster

    operator_scanner = re.compile(r'Operator         : (\S*)')
    time_scanner = re.compile(r'Time/Date        : (\d{2}:\d{2}:\d{2}\s*\w*\s*\d*,\s*\d*)')
    sampleid_scanner = re.compile(r'Sample ID        : (.*)')
    instrument_sn_scanner = re.compile(r'Instrument S/N   : (\S*)')

    collimatorefl_scanner = re.compile(r'EFL \(Collimator\): (\d*) mm')
    wavelength_scanner = re.compile(r'Wavelength      : (\d+) nm')
    sampleefl_scanner = re.compile(r'EFL \(Sample\)    : (\d*\.\d*) mm')
    objangle_scanner = re.compile(r'Object Angle    : (-?\d*\.\d*) =B0')
    focuspos_scanner = re.compile(r'Focus Position  : (\d*\.\d*) mm')
    azimuth_scanner = re.compile(r'Sample Azimuth  : (-?\d*\.\d*) =B0')

    operator = operator_scanner.search(data).group(1)
    time = time_scanner.search(data).group(1)
    hms, month, day, year = time.split()
    year, day = int(year), int(day[:-1])
    month_num = list(calendar.month_name).index(month)
    h, m, s = hms.split(':')
    h, m, s = (int(str_) for str_ in [h, m, s])
    timestamp = datetime.datetime(year=year, month=month_num, day=day, hour=h, minute=m, second=s)
    sampleid = sampleid_scanner.search(data).group(1).strip()
    instrument_sn = instrument_sn_scanner.search(data).group(1)

    collimator_efl = float(collimatorefl_scanner.search(data).group(1))
    wavelength = float(wavelength_scanner.search(data).group(1)) / 1e3  # nm to um
    sample_efl = float(sampleefl_scanner.search(data).group(1))
    obj_angle = float(objangle_scanner.search(data).group(1))
    focus_pos = float(focuspos_scanner.search(data).group(1))
    azimuth = float(azimuth_scanner.search(data).group(1))
    return {
        'operator': operator,
        'time': timestamp,
        'sample_id': sampleid,
        'instrument': 'Trioptics ImageMaster HR',
        'instrument_sn': instrument_sn,
        'collimator': collimator_efl,
        'wavelength': wavelength,
        'efl': sample_efl,
        'fno': None,
        'obj_angle': obj_angle,
        'focus_pos': focus_pos,
        'azimuth': azimuth,
    }


def identify_trioptics_measurement_type(file):
    """Identify type of measurement in a Trioptics .mht file.

    Parameters
    ----------
    file : `str` or path_like or file_like
        contents of a file, path_like to the file, or file object

    Returns
    -------
    program : `str`
        measurement type
    data : `str`
        contents of the file

    """
    data = read_file_stream_or_path(file)
    data_parse = data[750:1500]
    measurement_type_scanner = re.compile(r'Measure Program  : (.*)')
    program = measurement_type_scanner.search(data_parse).group(1).strip()
    return program, data


TRIOPTICS_SWITCHBOARD = {
    'MTF vs. Field': read_trioptics_mtf_vs_field,
    'Distortion': NotImplemented,
    'Axial Color': NotImplemented,
    'Lateral Color': NotImplemented,
}


def read_any_trioptics_mht(file, metadata=False):
    """Read any Trioptics .mht certificate (MTF vs Field, Distortion, etc).

    Parameters
    ----------
    file : `str` or path_like or file_like
        contents of a file, path_like to the file, or file object
    metadata : `bool`
        whether to also extract and return metadata

    Returns
    -------
    `dict`
        dictionary with appropriate keys.  If metadata=True, also has keys in
        the return of `io.parse_trioptics_metadata`.

    """
    type_, data = identify_trioptics_measurement_type(file)
    return type_, TRIOPTICS_SWITCHBOARD[type_](data, metadata=metadata)


def read_mtfmapper_sfr_single(file, pixel_pitch=None):
    """Read an MTF Mapper SFR (MTF) file generated by the -f flag with --single-roi.

    Notes
    -----
    This reads a "raw_sfr_values.txt" file, not an "edge_sfr_values.txt" file.

    Parameters
    ----------
    file : `str` or path_like or file_like
        contents of a file, path_like to the file, or file object
    pixel_pitch : `float`
        center-to-center pixel spacing, in microns

    Returns
    -------
    `numpy.ndarray`
        spatial_frequencies
    `numpy.ndarray`
        mtf

    """
    data = read_file_stream_or_path(file)
    floats = [float(d) for d in data.splitlines()[0].split(' ')[:-1]]
    edge_angle, *mtf = floats
    mtf = m.asarray(mtf)
    freqs = m.arange(len(mtf)) / 64
    if pixel_pitch is not None:  # convert cy/px to cy/mm
        freqs /= (pixel_pitch / 1e3)

    return freqs, mtf


def read_zygo_datx(file):
    """Read a zygo datx file.

    Parameters
    ----------
    file : path_like
        location of a file

    Returns
    -------
    `dict`
        dictionary with keys phase, intensity, meta

    Raises
    ------
    ImportError
        h5py unavailable, required dependency for this

    """
    import h5py

    # create a handle to the h5 file
    with h5py.File(file, 'r') as f:
        # cast intensity down to int16, saves memory and Zygo doesn't use cameras >> 16-bit
        try:
            intens_block = list(f['Data']['Intensity'].keys())[0]
            intensity = m.flipud(f['Data']['Intensity'][intens_block][()].astype(m.uint16))
        except KeyError:
            intensity = None
        phase_block = list(f['Data']['Surface'].keys())[0]
        phase = m.flipud(f['Data']['Surface'][phase_block][()])

        # now get attrs
        attrs = f['Attributes']
        key = list(attrs)[-1]
        attrs = attrs[key].attrs
        meta = {}
        for key, value in attrs.items():
            if key.endswith('Unit'):
                continue  # do not need unit keys, units implicitly understood.
            if '.Data Attributes.' in key:
                key = key.split('.Data Attributes.')[-1]
            if key.endswith('Value'):
                key = key[:-5]  # strip value from key
            if key.endswith(':'):
                key = key[:-1]
            if key == 'Resolution':
                key = 'Lateral Resolution'
            elif key.endswith('Data Context.Lateral Resolution'):
                continue  # duplicate
            elif key in ['Property Bag List', 'Group Number', 'TextCount']:
                continue  # h5py particulars
            if value.dtype == 'object':
                value = str(value[0])  # object dtype is a string
            elif value.dtype in ['uint8', 'int32']:
                value = int(value[0])
            elif value.dtype in ['float64']:
                value = float(value[0])
            else:
                continue  # compound items, h5py objects that do not map nicely to primitives

            meta[key] = value

        # lastly, obliquity factor.  Because Mx is spaghetti code and it can appear in many places
        # under different names, or not at all.  Thanks, Zygo.
        try:
            # "new style" datx files, Measurement is a branch of the h5 tree
            # with duplicates of surface and intenisty, but different attrs.
            # Pluck the obliquity from here, if possible.
            obliq = f['Measurement']['Surface'].attrs['Obliquity Factor']
        except KeyError:
            try:
                # possibly an "old style" datx file, "obliquity" in the master attrs
                obliq = (meta['Obliquity'],)
                del meta['Obliquity']
            except KeyError:
                obliq = (1,)

        meta['Obliquity Factor'] = float(obliq[0])

        was_nx2 = False

        # These may be Mx 7.3 and not NX2 problems, I don't know.
        if 'Interf Scale Factor' not in meta:
            # NX2-type datx file, look under surface attributes...
            meta['Interf Scale Factor'] = f['Measurement']['Surface'].attrs['Interferometric Scale Factor'][0]
            was_nx2 = True

        if 'Lateral Resolution' not in meta:
            # NX2-type.  magic numbers [0][2][1], h5 is such a great format...
            meta['Lateral Resolution'] = f['Measurement']['Surface'].attrs['X Converter'][0][2][1]
            was_nx2 = True

        if 'No Data' not in meta:
            try:
                meta['No Data'] = f['Measurement']['Surface'].attrs['No Data'][0]
            except KeyError:
                meta['No Data'] = ZYGO_INVALID_PHASE

    phase[phase >= meta['No Data']] = m.nan
    phase = phase.astype(config.precision)  # cast to big endian
    if not was_nx2:
        phase *= (meta['Interf Scale Factor'] * meta['Obliquity Factor'] * meta['Wavelength']) * 1e9

    return {
        'phase': phase,
        'intensity': intensity,
        'meta': meta,
    }


ZYGO_INVALID_PHASE = 2147483640
ZYGO_ENC = 'utf-8'  # may be ASCII, cp1252...
ZYGO_PHASE_RES_FACTORS = {
    0: 4096,
    1: 32768,
}


def read_zygo_dat(file, multi_intensity_action='first'):
    """Read the contents of a zygo binary (.dat) file.

    Parameters
    ----------
    file : path_like
        path to a file
    multi_itensity_action : `str`, {'avg', 'first', 'last'}
        action to take when handling multiple intensitiy frames, only avg is valid at this time

    Returns
    -------
    `dict`
        dictionary with keys: phase, intensity, meta

    """
    with open(file, 'rb') as fid:
        contents = fid.read()

    meta = read_zygo_metadata(contents)
    iw, ih, ib = meta['ac']['width'], meta['ac']['height'], meta['ac']['n_buckets']
    if ib == 0:
        ib = 1
    ilen = iw * ih * ib  # intensity
    pw, ph = meta['cn']['width'], meta['cn']['height']
    plen = pw * ph  # phase
    header_len = meta['header']['size']

    intensity = m.frombuffer(contents, offset=header_len, count=ilen, dtype=m.uint16).reshape((ib, ih, iw))
    if multi_intensity_action.lower() == 'avg':
        intensity = intensity.mean(axis=0)
    elif multi_intensity_action.lower() == 'first':
        intensity = intensity[0]
    elif multi_intensity_action.lower() == 'last':
        intensity = intensity[-1]
    else:
        raise ValueError(f'multi_intensity_action {multi_intensity_action} not among valid options of avg, first, last.')

    # little-endian camera data, not sure if always need to byteswap, may break for some users...
    phase_raw = m.frombuffer(contents, offset=header_len + ilen * 2, count=plen, dtype=m.int32)
    phase = phase_raw.copy().byteswap(True).astype(config.precision).reshape((ph, pw))
    phase[phase >= ZYGO_INVALID_PHASE] = m.nan
    phase *= (meta['scale_factor'] * meta['obliquity_factor'] * meta['wavelength'] /
              ZYGO_PHASE_RES_FACTORS[meta['phase_res']]) * 1e9  # unit m to nm
    return {
        'phase': phase,
        'intensity': intensity,
        'meta': meta,
    }


def read_zygo_metadata(file_contents):
    """Parse metadata from the contents of a binary Zygo file.

    Parameters
    ----------
    file_contents : `bytes`
        binary file contents

    Returns
    -------
    `dict`
        dictionary with a shitload of keys for all of Zygo's metadata.

    """
    # convenient single character name
    c = file_contents
    IB16 = '>H'
    IL16 = '<H'
    IB32 = '>I'
    IL32 = '<I'
    FB32 = '>f'
    FL32 = '<f'
    C = 'c'
    uint8 = 'B'
    WASTE_BYTE = '\x00'

    magic_number = struct.unpack(IB32, c[:4])[0]
    header = {
        'format': struct.unpack(IB16, c[4:6])[0],
        'size': struct.unpack(IB32, c[6:10])[0],
    }
    swtype = struct.unpack(IB16, c[10:12])[0]
    swdate = c[12:42].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    swmaj = struct.unpack(IB16, c[42:44])[0]
    swmin = struct.unpack(IB16, c[44:46])[0]
    swpatch = struct.unpack(IB16, c[46:48])[0]
    metropro_version = f'{swmaj}.{swmin}.{swpatch}'
    ac = {
        'x': struct.unpack(IB16, c[48:50])[0],
        'y': struct.unpack(IB16, c[50:52])[0],
        'width': struct.unpack(IB16, c[52:54])[0],
        'height': struct.unpack(IB16, c[54:56])[0],
        'n_buckets': struct.unpack(IB16, c[56:58])[0],
        'range': struct.unpack(IB16, c[58:60])[0],
        'n_bytes': struct.unpack(IB32, c[60:64])[0],
    }
    cn = {
        'x': struct.unpack(IB16, c[64:66])[0],
        'y': struct.unpack(IB16, c[66:68])[0],
        'width': struct.unpack(IB16, c[68:70])[0],
        'height': struct.unpack(IB16, c[70:72])[0],
        'n_bytes': struct.unpack(IB32, c[72:76])[0],
    }
    timestamp = struct.unpack(IB32, c[76:80])[0]
    comment = c[80:162].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    source = struct.unpack(IB16, c[162:164])[0]
    scale_factor = struct.unpack(FB32, c[164:168])[0]
    wavelength = struct.unpack(FB32, c[168:172])[0]
    numerical_aperture = struct.unpack(FB32, c[172:176])[0]
    obliquity_factor = struct.unpack(FB32, c[176:180])[0]
    magnification = struct.unpack(FB32, c[180:184])[0]
    lateral_resolution = struct.unpack(FB32, c[184:188])[0]
    acq_type = struct.unpack(IB16, c[188:190])[0]
    intensity_average_count = struct.unpack(IB16, c[190:192])[0]
    sfac_limit = struct.unpack(IB16, c[194:196])[0]
    ramp = {
        'cal': struct.unpack(IB16, c[192:194])[0],
        'gain': struct.unpack(IB16, c[196:198])[0],
    }
    part_thickness = struct.unpack(FB32, c[198:202])[0]
    sw_llc = struct.unpack(IB16, c[202:204])[0]
    target_range = struct.unpack(FB32, c[204:208])[0]
    rad_crv_measure_seq = struct.unpack(IL16, c[208:210])[0]
    min_mod = struct.unpack(IB32, c[210:214])[0]
    min_mod_count = struct.unpack(IB32, c[214:218])[0]
    phase_res = struct.unpack(IB16, c[218:220])[0]
    min_area = struct.unpack(IB32, c[220:224])[0]
    discontinuity = {
        'action': struct.unpack(IB16, c[224:226])[0],
        'filter': struct.unpack(FB32, c[226:230])[0],
    }
    connect_order = struct.unpack(IB16, c[230:232])[0]
    sign = struct.unpack(IB16, c[232:234])[0]
    camera = {
        'width': struct.unpack(IB16, c[234:236])[0],
        'height': struct.unpack(IB16, c[236:238])[0],
    }
    _sys = {
        'type': struct.unpack(IB16, c[238:240])[0],
        'board': struct.unpack(IB16, c[240:242])[0],
        'serial': struct.unpack(IB16, c[242:244])[0],
        'inst_id': struct.unpack(IB16, c[244:246])[0]
    }
    obj_name = c[246:258].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    part_name = c[258:298].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    codev_type = struct.unpack(IB16, c[298:300])[0]
    phase_avg_count = struct.unpack(IB16, c[300:302])[0]
    sub_sys_err = struct.unpack(IB16, c[302:304])[0]
    # 305-320 unused
    part_sn = c[320:360].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    refractive_index = struct.unpack(FB32, c[360:364])[0]
    remove = {
        'tilt': struct.unpack(IB16, c[364:366])[0],
        'fringes': struct.unpack(IB16, c[366:368])[0],
    }
    max_area = struct.unpack(IB32, c[368:372])[0]
    setup_type = struct.unpack(IB16, c[372:374])[0]
    wrapped = struct.unpack(IB16, c[374:376])[0]
    pre_connect_filter = struct.unpack(FB32, c[376:380])[0]
    wavelength_in = {
        1: struct.unpack(FB32, c[386:390])[0],
        2: struct.unpack(FB32, c[380:384])[0],
        3: struct.unpack(FB32, c[390:394])[0],
        4: struct.unpack(FB32, c[394:398])[0],
        'fold': struct.unpack(IB16, c[384:386])[0],
    }
    wavelength_select = c[398:406].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    fda_res = struct.unpack(IB16, c[406:408])[0]
    scan_description = c[408:428].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    # n_fiducials = struct.unpack(IB16, c[428:430])  # skip - redundant
    fiducials = [
        struct.unpack(FB32, c[430:434])[0],
        struct.unpack(FB32, c[434:438])[0],
        struct.unpack(FB32, c[438:442])[0],
        struct.unpack(FB32, c[442:446])[0],
        struct.unpack(FB32, c[446:450])[0],
        struct.unpack(FB32, c[450:454])[0],
        struct.unpack(FB32, c[454:458])[0],
        struct.unpack(FB32, c[458:462])[0],
        struct.unpack(FB32, c[462:466])[0],
        struct.unpack(FB32, c[466:470])[0],
        struct.unpack(FB32, c[470:474])[0],
        struct.unpack(FB32, c[474:478])[0],
        struct.unpack(FB32, c[478:482])[0],
        struct.unpack(FB32, c[482:486])[0],
    ]
    pixel_dims = {
        'width': struct.unpack(FB32, c[486:490])[0],
        'height': struct.unpack(FB32, c[490:494])[0]
    }
    exit_pupil_diameter = struct.unpack(FB32, c[494:498])[0]
    light_level_percent = struct.unpack(FB32, c[498:502])[0]
    coords = {
        'state': struct.unpack(IL32, c[502:506])[0],
        'x': struct.unpack(FL32, c[506:510])[0],
        'y': struct.unpack(FL32, c[510:514])[0],
        'z': struct.unpack(FL32, c[514:518])[0],
        'a': struct.unpack(FL32, c[518:522])[0],  # x rotation
        'b': struct.unpack(FL32, c[522:526])[0],  # y rotation
        'c': struct.unpack(FL32, c[526:530])[0],  # z rotation
    }
    coherence_mode = struct.unpack(IL16, c[530:532])[0]
    surface_filter = struct.unpack(IL16, c[532:534])[0]
    sys_err_filename = c[534:562].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    zoom_descr = c[562:570].decode(ZYGO_ENC).rstrip(WASTE_BYTE)
    # maybe can make a part dict, merge with above part_thickness, etc?
    alpha_part = struct.unpack(FL32, c[570:574])[0]
    beta_part = struct.unpack(FL32, c[574:578])[0]
    dist_part = struct.unpack(FL32, c[578:582])[0]
    cam_split = {
        'loc_x': struct.unpack(IL16, c[582:584])[0],
        'loc_y': struct.unpack(IL16, c[584:586])[0],
        'trans_x': struct.unpack(IL16, c[586:588])[0],
        'trans_y': struct.unpack(IL16, c[588:590])[0],
    }
    material = {
        'a': c[590:614].decode(ZYGO_ENC).rstrip(WASTE_BYTE),
        'b': c[614:638].decode(ZYGO_ENC).rstrip(WASTE_BYTE),
    }
    # 639-642 unused
    dmi_center = {
        'x': struct.unpack(FL32, c[642:646])[0],
        'y': struct.unpack(FL32, c[646:650])[0],
    }
    sph_distortion_correction = struct.unpack(IL16, c[650:652])[0]
    # 653-654 unused
    sph_dist = {
        'part_na': struct.unpack(FL32, c[654:658])[0],
        'part_radius': struct.unpack(FL32, c[658:662])[0],
        'cal_na': struct.unpack(FL32, c[662:666])[0],
        'cal_radius': struct.unpack(FL32, c[666:670])[0],
    }
    surface_type = struct.unpack(IL16, c[670:672])[0]
    ac_surface_type = struct.unpack(IL16, c[672:674])[0]
    z_pos = struct.unpack(FL32, c[674:678])[0]
    power_mul = struct.unpack(FL32, c[678:682])[0]
    focus_mul = struct.unpack(FL32, c[682:686])[0]
    roc_focus_cal_factor = struct.unpack(FL32, c[686:690])[0]
    roc_power_cal_factor = struct.unpack(FL32, c[690:694])[0]
    ftp_pos = {
        'left': struct.unpack(FL32, c[694:698])[0],
        'right': struct.unpack(FL32, c[698:702])[0],
        'pitch': struct.unpack(FL32, c[702:706])[0],
        'roll': struct.unpack(FL32, c[706:710])[0],
    }
    min_mod_percent = struct.unpack(FL32, c[710:714])[0]
    max_intens = struct.unpack(IL32, c[714:718])[0]
    ring_of_fire = struct.unpack(IL16, c[718:720])[0]  # lol wyd zygo
    # 721 unused
    rc = {
        'orientation': struct.unpack(C, c[721:722])[0].decode(ZYGO_ENC).rstrip(WASTE_BYTE),
        'distance': struct.unpack(FL32, c[722:726])[0],
        'angle': struct.unpack(FL32, c[726:730])[0],
        'diameter': struct.unpack(FL32, c[730:734])[0],
    }
    rem_fringes_mode = struct.unpack(IB16, c[734:736])[0]
    # 737 unused
    ftpsi_phase_res = struct.unpack(uint8, c[737:738])[0]
    frames_acquired = struct.unpack(IL16, c[738:740])[0]
    cavity_type = struct.unpack(IL16, c[740:742])[0]
    cam_frame_rate = struct.unpack(FL32, c[742:746])[0]
    tune_range = struct.unpack(FL32, c[746:750])[0]
    cal_pix = {
        'x': struct.unpack(IL16, c[750:752])[0],
        'y': struct.unpack(IL16, c[752:754])[0],
    }
    # n_test_cal_pts = struct.unpack(IL16, c[754:756])  # not bothering to read
    # n_ref_cal_pts = struct.unpack(IL16, c[756:758])   # these, redundant
    test_cal_pts = [
        struct.unpack(FL32, c[758:762])[0],
        struct.unpack(FL32, c[762:766])[0],
        struct.unpack(FL32, c[766:770])[0],
        struct.unpack(FL32, c[770:774])[0],
    ]
    ref_cal_pts = [
        struct.unpack(FL32, c[774:778])[0],
        struct.unpack(FL32, c[778:782])[0],
        struct.unpack(FL32, c[782:786])[0],
        struct.unpack(FL32, c[786:790])[0],
    ]
    test_cal_pix_opd = struct.unpack(FL32, c[790:794])[0]
    test_ref_pix_opd = struct.unpack(FL32, c[794:798])[0]
    flash_phase_cd_mask = struct.unpack(FL32, c[798:802])[0]
    flash_phase_alias_mask = struct.unpack(FL32, c[802:806])[0]
    flask_phase_filter = struct.unpack(FL32, c[806:810])[0]
    scan_direction = struct.unpack(uint8, c[810:811])[0]
    # 812 - 814 unused
    ftpsi_res_factor = struct.unpack(IL16, c[814:816])[0]
    # 835 - 900 films, for later
    # 901 - 4096 unused

    # combine distant vars
    scan = {
        'direction': scan_direction,
        'scan_description': scan_description,
    }
    all_vars = [
        magic_number,
        header,
        swtype,
        swdate,
        metropro_version,
        ac,
        cn,
        timestamp,
        comment,
        source,
        scale_factor,
        wavelength,
        numerical_aperture,
        obliquity_factor,
        magnification,
        lateral_resolution,
        acq_type,
        intensity_average_count,
        ramp,
        sfac_limit,
        part_thickness,
        sw_llc,
        target_range,
        rad_crv_measure_seq,
        min_mod,
        min_mod_count,
        phase_res,
        min_area,
        discontinuity,
        connect_order,
        sign,
        camera,
        _sys,
        obj_name,
        part_name,
        codev_type,
        phase_avg_count,
        sub_sys_err,
        part_sn,
        refractive_index,
        remove,
        max_area,
        setup_type,
        wrapped,
        pre_connect_filter,
        wavelength_in,
        wavelength_select,
        fda_res,
        scan,
        fiducials,
        pixel_dims,
        exit_pupil_diameter,
        light_level_percent,
        coords,
        coherence_mode,
        surface_filter,
        sys_err_filename,
        zoom_descr,
        alpha_part,
        beta_part,
        dist_part,
        cam_split,
        material,
        dmi_center,
        sph_distortion_correction,
        sph_dist,
        surface_type,
        ac_surface_type,
        z_pos,
        power_mul,
        focus_mul,
        roc_focus_cal_factor,
        roc_power_cal_factor,
        ftp_pos,
        min_mod_percent,
        max_intens,
        ring_of_fire,
        rc,
        rem_fringes_mode,
        ftpsi_phase_res,
        frames_acquired,
        cavity_type,
        cam_frame_rate,
        tune_range,
        cal_pix,
        ref_cal_pts,
        test_cal_pts,
        test_cal_pix_opd,
        test_ref_pix_opd,
        flash_phase_cd_mask,
        flash_phase_alias_mask,
        flask_phase_filter,
        ftpsi_res_factor,
    ]
    all_keys = [
        'magic_number',
        'header',
        'swtype',
        'swdate',
        'metropro_version',
        'ac',
        'cn',
        'timestamp',
        'comment',
        'source',
        'scale_factor',
        'wavelength',
        'numerical_aperture',
        'obliquity_factor',
        'magnification',
        'lateral_resolution',
        'acq_type',
        'intensity_average_count',
        'ramp',
        'sfac_limit',
        'part_thickness',
        'sw_llc',
        'target_range',
        'rad_crv_measure_seq',
        'min_mod',
        'min_mod_count',
        'phase_res',
        'min_area',
        'discontinuity',
        'connect_order',
        'sign',
        'camera',
        '_sys',
        'obj_name',
        'part_name',
        'codev_type',
        'phase_avg_count',
        'sub_sys_err',
        'part_sn',
        'refractive_index',
        'remove',
        'max_area',
        'setup_type',
        'wrapped',
        'pre_connect_filter',
        'wavelength_in',
        'wavelength_select',
        'fda_res',
        'scan',
        'fiducials',
        'pixel_dims',
        'exit_pupil_diameter',
        'light_level_percent',
        'coords',
        'coherence_mode',
        'surface_filter',
        'sys_err_filename',
        'zoom_descr',
        'alpha_part',
        'beta_part',
        'dist_part',
        'cam_split',
        'material',
        'dmi_center',
        'sph_distortion_correction',
        'sph_dist',
        'surface_type',
        'ac_surface_type',
        'z_pos',
        'power_mul',
        'focus_mul',
        'roc_focus_cal_factor',
        'roc_power_cal_factor',
        'ftp_pos',
        'min_mod_percent',
        'max_intens',
        'ring_of_fire',
        'rc',
        'rem_fringes_mode',
        'ftpsi_phase_res',
        'frames_acquired',
        'cavity_type',
        'cam_frame_rate',
        'tune_range',
        'cal_pix',
        'ref_cal_pts',
        'test_cal_pts',
        'test_cal_pix_opd',
        'test_ref_pix_opd',
        'flash_phase_cd_mask',
        'flash_phase_alias_mask',
        'flask_phase_filter',
        'ftpsi_res_factor',
    ]

    return {k: v for k, v in zip(all_keys, all_vars)}



"""phase basics."""

from .util import share_fig_ax, pv, rms, Sa, std

from prysm import mathops as m


class OpticalPhase(object):
    units = {
        'm': 'm',
        'meter': 'm',
        'mm': 'mm',
        'millimeter': 'mm',
        'μm': 'μm',
        'um': 'μm',
        'micron': 'μm',
        'micrometer': 'μm',
        'nm': 'nm',
        'nanometer': 'nm',
        'Å': 'Å',
        'aa': 'Å',
        'angstrom': 'Å',
        'λ': 'λ',
        'waves': 'λ',
        'lambda': 'λ',
        'px': 'px',
        'pixel': 'px',
    }
    unit_scales = {
        'm': 1,
        'mm': 1e-3,
        'μm': 1e-6,
        'nm': 1e-9,
        'Å': 1e-10,
    }
    unit_changes = {
        'm_m': lambda x: 1,
        'm_mm': lambda x: 1e-3,
        'm_μm': lambda x: 1e-6,
        'm_nm': lambda x: 1e-9,
        'm_Å': lambda x: 1e-10,
        'm_λ': lambda x: 1e-6 / x,
        'mm_mm': lambda x: 1,
        'mm_m': lambda x: 1e3,
        'mm_μm': lambda x: 1e-3,
        'mm_nm': lambda x: 1e-6,
        'mm_Å': lambda x: 1e-7,
        'mm_λ': lambda x: 1e-3 / x,
        'μm_μm': lambda x: 1,
        'μm_m': lambda x: 1e6,
        'μm_mm': lambda x: 1e3,
        'μm_nm': lambda x: 1e-3,
        'μm_Å': lambda x: 1e-4,
        'μm_λ': lambda x: 1 / x,
        'nm_nm': lambda x: 1,
        'nm_m': lambda x: 1e9,
        'nm_mm': lambda x: 1e6,
        'nm_μm': lambda x: 1e3,
        'nm_Å': lambda x: 1e-1,
        'nm_λ': lambda x: 1e3 / x,
        'Å_Å': lambda x: 1,
        'Å_m': lambda x: 1e10,
        'Å_mm': lambda x: 1e7,
        'Å_μm': lambda x: 1e4,
        'Å_nm': lambda x: 10,
        'Å_λ': lambda x: 1e4 / x,
        'λ_λ': lambda x: 1,
        'λ_m': lambda x: 1e6 * x,
        'λ_mm': lambda x: 1e3 * x,
        'λ_um': lambda x: x,
        'λ_nm': lambda x: 1e-3 * x,
        'λ_Å': lambda x: 1e-4 * x,
        'px_px': lambda x: 1,  # beware changing pixels to other units
        'px_m': lambda x: 1,
        'px_mm': lambda x: 1,
        'px_μm': lambda x: 1,
        'px_nm': lambda x: 1,
        'px_Å': lambda x: 1,
        'px_λ': lambda x: 1,
    }

    def __init__(self, unit_x, unit_y, phase, phase_unit, spatial_unit, wavelength=None):
        """Create a new instance of an OpticalPhase.

        Note that this class is not intended to be used directly, and is meant
        to allow shared functionality and interchange between the `Pupil` and
        `Interferogram` classes.

        Parameters
        ----------
        unit_x : `np.ndarray`
            x spatial units
        unit_y : `np.ndarray`
            y spatial units
        phase : `np.ndarray`
            phase/height/opd data
        phase_unit : `str`
            unit used to describe the phase, see `OpticalPhase`.units
        spatial_unit : `str`
            unit used to describe unit_x and unit_y, see `OpticalPhase`.units
        wavelength : `float`, optional
            wavelength of light, in microns

        """
        self.unit_x = unit_x
        self.unit_y = unit_y
        self.phase = phase
        self.wavelength = wavelength
        self.phase_unit = phase_unit
        self.spatial_unit = spatial_unit
        self.xaxis_label = 'x'  # these should be overriden by subclasses or instances
        self.yaxis_label = 'y'
        self.zaxis_label = 'z'

    @property
    def phase_unit(self):
        return self._phase_unit

    @phase_unit.setter
    def phase_unit(self, unit):
        unit = unit.lower()
        if unit == 'å':
            self._phase_unit = unit.upper()
        else:
            if unit not in self.units:
                raise ValueError(f'{unit} not a valid unit, must be in {set(self.units.keys())}')
            self._phase_unit = self.units[unit]

    @property
    def spatial_unit(self):
        return self._spatial_unit

    @spatial_unit.setter
    def spatial_unit(self, unit):
        unit = unit.lower()
        if unit not in self.units:
            raise ValueError(f'{unit} not a valid unit, must be in {set(self.units.keys())}')

        self._spatial_unit = self.units[unit]

    @property
    def slice_x(self):
        """Retrieve a slice through the X axis of the phase.

        Returns
        -------
        self.unit : `numpy.ndarray`
            ordinate axis
        slice of self.phase : `numpy.ndarray`

        """
        return self.unit_x, self.phase[self.center_y, :]

    @property
    def slice_y(self):
        """Retrieve a slice through the Y axis of the phase.

        Returns
        -------
        self.unit : `numpy.ndarray`
            ordinate axis
        slice of self.phase : `numpy.ndarray`

        """
        return self.unit_y, self.phase[:, self.center_x]

    @property
    def pv(self):
        """Peak-to-Valley phase error.  DIN/ISO 'St.'"""
        return pv(self.phase)

    @property
    def rms(self):
        """RMS phase error.  DIN/ISO 'Sq.'"""
        return rms(self.phase)

    @property
    def Sa(self):
        """Sa phase error.  DIN/ISO 'Sa.'"""
        return Sa(self.phase)

    @property
    def std(self):
        """Standard deviation of phase error."""
        return std(self.phase)

    @property
    def shape(self):
        """Proxy to self.phase.shape."""
        return self.phase.shape

    @property
    def diameter_x(self):
        """Diameter of the data in x."""
        return self.unit_x[-1] - self.unit_x[0]

    @property
    def diameter_y(self):
        """Diameter of the data in y."""
        return self.unit_y[-1] - self.unit_x[0]

    @property
    def diameter(self):
        """Greater of (self.diameter_x, self.diameter_y)."""
        return max((self.diameter_x, self.diameter_y))

    @property
    def semidiameter(self):
        """Half of self.diameter."""
        return self.diameter / 2

    @property
    def samples_x(self):
        """Number of samples in the x dimension."""
        return self.phase.shape[1]

    @property
    def samples_y(self):
        """Number of samples in the y dimension."""
        return self.phase.shape[0]

    @property
    def sample_spacing(self):
        """center-to-center sample spacing."""
        return self.unit_x[1] - self.unit_x[0]

    @property
    def center_x(self):
        """Center "pixel" in x."""
        return len(self.unit_x) // 2

    @property
    def center_y(self):
        """Center "pixel" in y."""
        return len(self.unit_y) // 2

    def change_phase_unit(self, to, inplace=True):
        """Change the units used to describe the phase.

        Parameters
        ----------
        to : `str`
            new unit, a member of `OpticalPhase`.units.keys()
        inplace : `bool`, optional
            whether to change self.phase, if False, returns updated phase, if True, returns self.

        Returns
        -------
        `new_phase` : `np.ndarray`
            new phase data
        OR
        `self` : `OpticalPhase`
            self
        """
        fctr = self.unit_changes['_'.join([self.phase_unit, self.units[to]])](self.wavelength)
        new_phase = self.phase / fctr
        if inplace:
            self.phase = new_phase
            self.phase_unit = to
            return self
        else:
            return new_phase

    def change_spatial_unit(self, to, inplace=True):
        """Change the units used to describe the spatial dimensions.

        Parameters
        ----------
        to : `str`
            new unit, a member of `OpticalPhase`.units.keys()
        inplace : `bool`, optional
            whether to change self.unit_x and self.unit_y.
            If False, returns updated phase, if True, returns self

        Returns
        -------
        `new_ux` : `np.ndarray`
            new ordinate x axis
        `new_uy` : `np.ndarray`
            new ordinate y axis
        OR
        `self` : `OpticalPhase`
            self

        """
        fctr = self.unit_changes['_'.join([self.spatial_unit, self.units[to]])](self.wavelength)
        new_ux = self.unit_x / fctr
        new_uy = self.unit_y / fctr
        if inplace:
            self.unit_x = new_ux
            self.unit_y = new_uy
            self.spatial_unit = to
            return self
        else:
            return new_ux, new_uy

    def plot2d(self, cmap='inferno', clim=(None, None), interp_method='lanczos', show_colorbar=True, fig=None, ax=None):
        """Plot the phase in 2D.

        Parameters
        ----------
        cmap : `str`
            colormap to use, passed directly to matplotlib
        interp_method : `str`, optional
            interpolation method to use, passed directly to matplotlib
        show_colorbar : `bool`, optional
            whether to draw the colorbar
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        fig, ax = share_fig_ax(fig, ax)

        if clim and not hasattr(clim, '__iter__'):
            clim = (-clim, clim)

        im = ax.imshow(self.phase,
                       extent=[self.unit_x[0], self.unit_x[-1], self.unit_y[0], self.unit_y[-1]],
                       cmap=cmap,
                       clim=clim,
                       origin='lower',
                       interpolation=interp_method)

        if show_colorbar:
            fig.colorbar(im, label=f'{self.zaxis_label} [{self.phase_unit}]', ax=ax, fraction=0.046)

        xlab = f'{self.xaxis_label} [{self.spatial_unit}]'
        ylab = f'{self.yaxis_label} [{self.spatial_unit}]'
        ax.set(xlabel=xlab, ylabel=ylab)
        return fig, ax

    def plot_slice_xy(self, fig=None, ax=None):
        """Create a plot of slices through the X and Y axes of the `Pupil`.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`, optional
            Figure to draw plot in
        ax : `matplotlib.axes.Axis`
            Axis to draw plot in

        Returns
        -------
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        """
        ux, x = self.slice_x
        uy, y = self.slice_y

        fig, ax = share_fig_ax(fig, ax)

        ax.plot(ux, x, lw=3, label='Slice X')
        ax.plot(uy, y, lw=3, label='Slice Y')
        ax.set(xlabel=f'{self.xaxis_label} [{self.spatial_unit}]',
               ylabel=f'{self.zaxis_label} [{self.phase_unit}]')
        ax.legend()
        return fig, ax

    def interferogram(self, visibility=1, passes=2, interp_method='lanczos', fig=None, ax=None):
        """Create an interferogram of the `Pupil`.

        Parameters
        ----------
        visibility : `float`
            Visibility of the interferogram
        passes : `float`
            Number of passes (double-pass, quadra-pass, etc.)
        interp_method : `str`, optional
            interpolation method, passed directly to matplotlib
        fig : `matplotlib.figure.Figure`, optional
            Figure to draw plot in
        ax : `matplotlib.axes.Axis`
            Axis to draw plot in

        Returns
        -------
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional:
            Axis containing the plot

        """
        epd = self.diameter
        phase = self.change_phase_unit(to='waves', inplace=False)

        fig, ax = share_fig_ax(fig, ax)
        plotdata = (visibility * m.sin(2 * m.pi * passes * phase))
        im = ax.imshow(plotdata,
                       extent=[-epd / 2, epd / 2, -epd / 2, epd / 2],
                       cmap='Greys_r',
                       interpolation=interp_method,
                       clim=(-1, 1),
                       origin='lower')
        fig.colorbar(im, label=r'Wrapped Phase [$\lambda$]', ax=ax, fraction=0.046)
        ax.set(xlabel=r'Pupil $\xi$ [mm]',
               ylabel=r'Pupil $\eta$ [mm]')
        return fig, ax



"""tools to analyze interferometric data."""
import warnings

from .conf import config
from ._phase import OpticalPhase
from ._zernike import defocus
from .io import read_zygo_dat, read_zygo_datx
from .fttools import forward_ft_unit
from .coordinates import cart_to_polar, uniform_cart_to_polar
from .propagation import prop_pupil_plane_to_psf_plane
from .util import share_fig_ax
from .geometry import mcache


from prysm import mathops as m


class Interferogram(OpticalPhase):
    """Class containing logic and data for working with interferometric data."""
    def __init__(self, phase, intensity=None, x=None, y=None, scale='px', phase_unit='nm', meta=None):
        if x is None:  # assume x, y given together
            x = m.arange(phase.shape[1])
            y = m.arange(phase.shape[0])
            scale = 'px'
            self.lateral_res = 1

        wvl = meta.get('wavelength', None)
        if wvl is None:
            wvl = meta.get('Wavelength')

        if wvl is not None:
            wvl *= 1e6  # m to um

        super().__init__(unit_x=x, unit_y=y, phase=phase,
                         wavelength=wvl, phase_unit=phase_unit,
                         spatial_unit=scale)

        self.xaxis_label = 'X'
        self.yaxis_label = 'Y'
        self.zaxis_label = 'Height'
        self.intensity = intensity
        self.meta = meta

        if scale != 'px':
            self.change_spatial_unit(to=scale, inplace=True)

    @property
    def dropout_percentage(self):
        """Percentage of pixels in the data that are invalid (NaN)."""
        return m.count_nonzero(~m.isfinite(self.phase)) / self.phase.size * 100

    @property
    def pvr(self):
        """Peak-to-Valley residual.

        Notes
        -----

        See:
        C. Evans, "Robust Estimation of PV for Optical Surface Specification and Testing"
        in Optical Fabrication and Testing, OSA Technical Digest (CD)
        (Optical Society of America, 2008), paper OWA4.
        http://www.opticsinfobase.org/abstract.cfm?URI=OFT-2008-OWA4

        """
        from .fringezernike import fit, FringeZernike
        coefs, residual = fit(self.phase, terms=36, residual=True)
        fz = FringeZernike(coefs, samples=self.shape[0])
        return fz.pv + 3 * residual

    def fill(self, _with=0):
        """Fill invalid (NaN) values.

        Parameters
        ----------
        _with : `float`, optional
            value to fill with

        Returns
        -------
        `Interferogram`
            self

        """
        nans = ~m.isfinite(self.phase)
        self.phase[nans] = _with
        return self

    def crop(self):
        """Crop data to rectangle bounding non-NaN region."""
        nans = m.isfinite(self.phase)
        nancols = m.any(nans, axis=0)
        nanrows = m.any(nans, axis=1)

        left, right = nanrows.argmax(), nanrows[::-1].argmax()
        top, bottom = nancols.argmax(), nancols[::-1].argmax()
        if left == right == top == bottom == 0:
            return self

        self.phase = self.phase[left:-right, top:-bottom]
        self.unit_y, self.unit_x = self.unit_y[left:-right], self.unit_x[top:-bottom]
        self.unit_x -= self.unit_x[0]
        self.unit_y -= self.unit_y[0]
        return self

    def remove_piston(self):
        """Remove piston from the data by subtracting the mean value."""
        self.phase -= self.phase[m.isfinite(self.phase)].mean()
        return self

    def remove_tiptilt(self):
        """Remove tip/tilt from the data by least squares fitting and subtracting a plane."""
        plane = fit_plane(self.unit_x, self.unit_y, self.phase)
        self.phase -= plane
        return self

    def remove_power(self):
        """Remove power from the data by least squares fitting."""
        sphere = fit_sphere(self.phase)
        self.phase -= sphere
        return self

    def remove_piston_tiptilt(self):
        """Remove piston/tip/tilt from the data, see remove_tiptilt and remove_piston."""
        self.remove_piston()
        self.remove_tiptilt()
        return self

    def remove_piston_tiptilt_power(self):
        """Remove piston/tip/tilt/power from the data."""
        self.remove_piston()
        self.remove_piston_tiptilt()
        self.remove_power()
        return self

    def mask(self, shape=None, diameter=0, mask=None):
        """Mask the signal.

        The mask will be inscribed in the axis with fewer pixels.  I.e., for
        a interferogram with 1280x1000 pixels, the mask will be 1000x1000 at
        largest.

        Parameters
        ----------
        shape : `str`
            valid shape from prysm.geometry
        diameter : `float`
            diameter of the mask, in self.spatial_units
        mask : `numpy.ndarray`
            user-provided mask

        Returns
        -------
        self
            modified Interferogram instance.

        """
        if shape is None and mask is None:
            raise ValueError('must provide either a shape or a mask')

        if mask is None:
            mask = mcache(shape, min(self.shape), radius=diameter / min(self.diameter_x, self.diameter_y))
            base = m.zeros(self.shape, dtype=config.precision)
            difference = abs(self.shape[0] - self.shape[1])
            l, u = int(m.floor(difference / 2)), int(m.ceil(difference / 2))
            if u is 0:  # guard against nocrop scenario
                _slice = slice(None)
            else:
                _slice = slice(l, -u)
            if self.shape[0] < self.shape[1]:
                base[:, _slice] = mask
            else:
                base[_slice, :] = mask

            mask = base

        hitpts = mask == 0
        self.phase[hitpts] = m.nan
        return self

    def latcal(self, plate_scale, unit='mm'):
        """Perform lateral calibration.

        This probably won't do what you want if your data already has spatial
        units of anything but pixels (px).

        Parameters
        ----------
        plate_scale : `float`
            center-to-center sample spacing of pixels, in (unit)s.
        unit : `str`, optional
            unit associated with the plate scale.

        Returns
        -------
        self
            modified `Interferogram` instance.

        """
        self.change_spatial_unit(to=unit, inplace=True)  # will be 0..n spatial units
        # sloppy to do this here...
        self.unit_x *= plate_scale
        self.unit_y *= plate_scale
        return self

    def spike_clip(self, nsigma=3):
        """Clip points in the data that exceed a certain multiple of the standard deviation.

        Parameters
        ----------
        nsigma : `float`
            number of standard deviations to keep

        Returns
        -------
        self
            this Interferogram instance.

        """
        pts_over_nsigma = abs(self.phase) > nsigma * self.std
        self.phase[pts_over_nsigma] = m.nan
        return self

    def psd(self):
        """Power spectral density of the data., units (self.phase_unit^2)/((cy/self.spatial_unit)^2).

        Returns
        -------
        unit_x : `numpy.ndarray`
            ordinate x frequency axis
        unit_y : `numpy.ndarray`
            ordinate y frequency axis
        psd : `numpy.ndarray`
            power spectral density

        """
        return psd(self.phase, self.sample_spacing)

    def bandlimited_rms(self, wllow=None, wlhigh=None, flow=None, fhigh=None):
        """Calculate the bandlimited RMS of a signal from its PSD.

        Parameters
        ----------
        wllow : `float`
            short spatial scale
        wlhigh : `float`
            long spatial scale
        flow : `float`
            low frequency
        fhigh : `float`
            high frequency

        Returns
        -------
        `float`
            band-limited RMS value.

        """
        return bandlimited_rms(*self.psd(),
                               wllow=wllow,
                               wlhigh=wlhigh,
                               flow=flow,
                               fhigh=fhigh)

    def total_integrated_scatter(self, wavelength, incident_angle=0):
        """Calculate the total integrated scatter (TIS) for an angle or angles.

        Parameters
        ----------
        wavelength : `float`
            wavelength of light in microns.
        incident_angle : `float` or `numpy.ndarray`
            incident angle(s) of light.

        Returns
        -------
        `float` or `numpy.ndarray`
            TIS value.

        """
        if self.spatial_unit != 'μm':
            raise ValueError('Use microns for spatial unit when evaluating TIS.')

        upper_limit = 1 / wavelength
        kernel = 4 * m.pi * m.cos(m.radians(incident_angle))
        kernel *= self.bandlimited_rms(upper_limit, None) / wavelength
        return 1 - m.exp(-kernel**2)

    def psd_xy_avg(self):
        """Power spectral density of the data., units (self.phase_unit^2)/((cy/self.spatial_unit)^2).

        Returns
        -------
        `dict`
            with keys x, y, avg.  Each containing a tuple of (unit, psd)

        """
        x, y, _psd = self.psd()
        lx, ly = len(x)//2, len(y)//2

        rho, phi, _psdrp = uniform_cart_to_polar(x, y, _psd)

        return {
            'x': (x[lx:], _psd[ly, lx:]),
            'y': (y[ly:], _psd[ly:, lx]),
            'avg': (rho, _psdrp.mean(axis=0)),
        }

    def plot_psd2d(self, axlim=None, clim=(1e-9, 1e2), interp_method='lanczos', fig=None, ax=None):
        """Plot the two dimensional PSD.

        Parameters
        ----------
        axlim : `float`, optional
            symmetrical axis limit
        power : `float`, optional
            inverse of power to stretch image by
        interp_method : `str`, optional
            method used to interpolate the image, passed directly to matplotlib
            imshow
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import colors
        x, y, psd = self.psd()

        if axlim is None:
            lims = (None, None)
        else:
            lims = (-axlim, axlim)

        fig, ax = share_fig_ax(fig, ax)
        im = ax.imshow(psd,
                       extent=[x[0], x[-1], y[0], y[-1]],
                       origin='lower',
                       cmap='Greys_r',
                       norm=colors.LogNorm(*clim),
                       interpolation=interp_method)

        ax.set(xlim=lims, xlabel=r'$\nu_x$' + f' [cy/{self.spatial_unit}]',
               ylim=lims, ylabel=r'$\nu_y$' + f' [cy/{self.spatial_unit}]')

        cb = fig.colorbar(im,
                          label='PSD [' + self.phase_unit + r'$^2$' + f'/(cy/{self.spatial_unit})]',
                          ax=ax, fraction=0.046, extend='both')
        cb.outline.set_edgecolor('k')
        cb.outline.set_linewidth(0.5)

        return fig, ax

    def plot_psd_xy_avg(self, a=None, b=None, c=None, lw=3,
                       xlim=None, ylim=None, fig=None, ax=None):
        """Plot the x, y, and average PSD on a linear x axis.

        Parameters
        ----------
        a : `float`, optional
            a coefficient of Lorentzian PSD model plotted alongside data
        b : `float`, optional
            b coefficient of Lorentzian PSD model plotted alongside data
        c : `float`, optional
            c coefficient of Lorentzian PSD model plotted alongside data
        lw : `float`, optional
            linewidth provided directly to matplotlib
        xlim : `tuple`, optional
            len 2 tuple of low, high x axis limits
        ylim : `tuple`, optional
            len 2 tuple of low, high y axis limits
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Notes
        -----
        if a, b given but not c, an AB / inverse power model will be used for the PSD.
        If a, b, c are given the Lorentzian model will be used.

        """
        xyavg = self.psd_xy_avg()
        x, px = xyavg['x']
        y, py = xyavg['y']
        r, pr = xyavg['avg']

        fig, ax = share_fig_ax(fig, ax)
        ax.loglog(x, px, lw=lw, label='x', alpha=0.4)
        ax.loglog(y, py, lw=lw, label='y', alpha=0.4)
        ax.loglog(r, pr, lw=lw*1.5, label='avg')

        if a is not None:
            if c is not None:
                requirement = abc_psd(a=a, b=b, c=c, nu=r)
            else:
                requirement = ab_psd(a=a, b=b, nu=r)
            ax.loglog(r, requirement, c='k', lw=lw*2)

        ax.legend(title='Orientation')
        ax.set(xlim=xlim, xlabel=f'Spatial Frequency [cy/{self.spatial_unit}]',
               ylim=ylim, ylabel=r'PSD [nm$^2$/' + f'(cy/{self.spatial_unit})$^2$]')

        return fig, ax

    @staticmethod
    def from_zygo_dat(path, multi_intensity_action='first', scale='mm'):
        """Create a new interferogram from a zygo dat file.

        Parameters
        ----------
        path : path_like
            path to a zygo dat file
        multi_intensity_action : str, optional
            see `io.read_zygo_dat`
        scale : `str`, optional, {'um', 'mm'}
            what xy scale to label the data with, microns or mm

        Returns
        -------
        `Interferogram`
            new Interferogram instance

        """
        if str(path).endswith('datx'):
            # datx file, use datx reader
            zydat = read_zygo_datx(path)
            res = zydat['meta']['Lateral Resolution']
        else:
            # dat file, use dat file reader
            zydat = read_zygo_dat(path, multi_intensity_action=multi_intensity_action)
            res = zydat['meta']['lateral_resolution']  # meters

        phase = zydat['phase']
        if res == 0.0:
            res = 1
            scale = 'px'

        if scale != 'px':
            _scale = 'm'
        else:
            _scale = 'px'

        i = Interferogram(phase=phase, intensity=zydat['intensity'],
                          x=m.arange(phase.shape[1]) * res, y=m.arange(phase.shape[0]) * res,
                          scale=_scale, meta=zydat['meta'])
        return i.change_spatial_unit(to=scale.lower(), inplace=True)


def fit_plane(x, y, z):
    xx, yy = m.meshgrid(x, y)
    pts = m.isfinite(z)
    xx_, yy_ = xx[pts].flatten(), yy[pts].flatten()
    flat = m.ones(xx_.shape)

    coefs = m.lstsq(m.stack([xx_, yy_, flat]).T, z[pts].flatten(), rcond=None)[0]
    plane_fit = coefs[0] * xx + coefs[1] * yy + coefs[2]
    return plane_fit


def fit_sphere(z):
    x, y = m.linspace(-1, 1, z.shape[1]), m.linspace(-1, 1, z.shape[0])
    xx, yy = m.meshgrid(x, y)
    pts = m.isfinite(z)
    xx_, yy_ = xx[pts].flatten(), yy[pts].flatten()
    rho, phi = cart_to_polar(xx_, yy_)
    focus = defocus(rho, phi)

    coefs = m.lstsq(m.stack([focus, m.ones(focus.shape)]).T, z[pts].flatten(), rcond=None)[0]
    rho, phi = cart_to_polar(xx, yy)
    sphere = defocus(rho, phi) * coefs[0]
    return sphere


def make_window(signal, sample_spacing, which='welch'):
    """Generates a window function to be used in PSD analysis.

    Parameters
    ----------
    signal : `numpy.ndarray`
        signal or phase data
    sample_spacing : `float`
        spacing of samples in the input data
    which : `str,` {'welch', 'hann', 'auto'}, optional
        which window to produce.  If auto, attempts to guess the appropriate
        window based on the input signal

    Notes
    -----
    For 2D welch, see:
    Power Spectral Density Specification and Analysis of Large Optical Surfaces
    E. Sidick, JPL

    Returns
    -------
    `numpy.ndarray`
        window array

    """
    s = signal.shape

    if which is None:
        # attempt to guess best window
        ysamples = int(round(s[0] * 0.02, 0))
        xsamples = int(round(s[1] * 0.02, 0))
        corner1 = signal[:ysamples, :xsamples] == 0
        corner2 = signal[-ysamples:, :xsamples] == 0
        corner3 = signal[:ysamples, -xsamples:] == 0
        corner4 = signal[-ysamples:, -xsamples:] == 0
        if corner1.all() and corner2.all() and corner3.all() and corner4.all():
            # four corners all "black" -- circular data, Welch window is best
            # looks wrong but 2D welch takes x, y while indices are y, x
            y = m.arange(s[1]) * sample_spacing
            x = m.arange(s[0]) * sample_spacing
            return window_2d_welch(y, x)
        else:
            # if not circular, square data; use Hanning window
            y = m.hanning(s[0])
            x = m.hanning(s[1])
            return m.outer(y, x)
    else:
        if type(which) is str:
            # known window type
            wl = which.lower()
            if wl == 'welch':
                y = m.arange(s[1]) * sample_spacing
                x = m.arange(s[0]) * sample_spacing
                return window_2d_welch(y, x)
            elif wl in ('hann', 'hanning'):
                y = m.hanning(s[0])
                x = m.hanning(s[1])
                return m.outer(y, x)
            else:
                raise ValueError('unknown window type')
        else:
            return which  # window provided as ndarray


def psd(height, sample_spacing, window=None):
    """Compute the power spectral density of a signal.

    Parameters
    ----------
    height : `numpy.ndarray`
        height or phase data
    sample_spacing : `float`
        spacing of samples in the input data
    window : {'welch', 'hann'} or ndarray, optional

    Returns
    -------
    unit_x : `numpy.ndarray`
        ordinate x frequency axis
    unit_y : `numpy.ndarray`
        ordinate y frequency axis
    psd : `numpy.ndarray`
        power spectral density

    Notes
    -----
    See GH_FFT for a rigorous treatment of FFT scalings
    https://holometer.fnal.gov/GH_FFT.pdf

    """
    window = make_window(height, sample_spacing, window)
    fft = m.ifftshift(m.fft2(m.fftshift(height * window)))
    psd = abs(fft)**2  # mag squared first as per GH_FFT

    fs = 1 / sample_spacing
    S2 = (window**2).sum()
    coef = S2 * fs * fs
    psd /= coef

    ux = forward_ft_unit(sample_spacing, height.shape[1])
    uy = forward_ft_unit(sample_spacing, height.shape[0])
    return ux, uy, psd


def bandlimited_rms(ux, uy, psd, wllow=None, wlhigh=None, flow=None, fhigh=None):
    """Calculate the bandlimited RMS of a signal from its PSD.

    Parameters
    ----------
    ux : `numpy.ndarray`
        x spatial frequencies
    uy : `numpy.ndarray`
        y spatial frequencies
    psd : `numpy.ndarray`
        power spectral density
    wllow : `float`
        short spatial scale
    wlhigh : `float`
        long spatial scale
    flow : `float`
        low frequency
    fhigh : `float`
        high frequency

    Returns
    -------
    `float`
        band-limited RMS value.

    """
    if wllow is not None or wlhigh is not None:
        # spatial period given
        if wllow is None:
            flow = 0
        else:
            fhigh = 1 / wllow

        if wlhigh is None:
            fhigh = max(ux[-1], uy[-1])
        else:
            flow = 1 / wlhigh
    elif flow is not None or fhigh is not None:
        # spatial frequency given
        if flow is None:
            flow = 0
        if fhigh is None:
            fhigh = max(ux[-1], uy[-1])
    else:
        raise ValueError('must specify either period (wavelength) or frequency')

    ux2, uy2 = m.meshgrid(ux, uy)
    r, p = cart_to_polar(ux2, uy2)

    if flow is None:
        warnings.warn('no lower limit given, using 0 for low frequency')
        flow = 0

    if fhigh is None:
        warnings.warn('no upper limit given, using limit imposed by data.')
        fhigh = r.max()

    work = psd.copy()
    work[r < flow] = 0
    work[r > fhigh] = 0
    first = m.trapz(work, uy, axis=0)
    second = m.trapz(first, ux, axis=0)
    return m.sqrt(second)


def window_2d_welch(x, y, alpha=8):
    """Return a 2D welch window for a given alpha.

    Parameters
    ----------
    x : `numpy.ndarray`
        x values, 1D array
    y : `numpy.ndarray`
        y values, 1D array
    alpha : `float`
        alpha (edge roll) parameter

    Returns
    -------
    `numpy.ndarray`
        window

    """
    xx, yy = m.meshgrid(x, y)
    r, _ = cart_to_polar(xx, yy)
    rmax = m.sqrt(x.max()**2 + y.max()**2)
    window = 1 - abs(r/rmax)**alpha
    return window


def abc_psd(nu, a, b, c):
    """Lorentzian model of a Power Spectral Density.

    Parameters
    ----------
    nu : `numpy.ndarray` or `float`
        spatial frequency
    a : `float`
        a coefficient
    b : `float`
        b coefficient
    c : `float`
        c coefficient

    Returns
    -------
    `numpy.ndarray`
        value of PSD model

    """
    return a / (1 + (nu/b)**2)**(c/2)


def ab_psd(nu, a, b):
    """inverse power model of a Power Spectral Density.

    Parameters
    ----------
    nu : `numpy.ndarray` or `float`
        spatial frequency
    a : `float`
        a coefficient
    b : `float`
        b coefficient

    Returns
    -------
    `numpy.ndarray`
        value of PSD model

    """
    return a * nu ** (-b)



"""prysm, a python optics module."""

from prysm.conf import config
from prysm.convolution import Convolvable
from prysm.extras import plot_fourier_chain
from prysm.detector import Detector, OLPF, PixelAperture
from prysm.pupil import Pupil
from prysm.fringezernike import FringeZernike
from prysm.standardzernike import StandardZernike
from prysm.seidel import Seidel
from prysm.psf import PSF, AiryDisk
from prysm.otf import MTF
from prysm.interferogram import Interferogram
from prysm.geometry import (
    circle,
    truecircle,
    gaussian,
    rotated_ellipse,
    square,
    regular_polygon,
    pentagon,
    hexagon,
    heptagon,
    octagon,
    nonagon,
    decagon,
    hendecagon,
    dodecagon,
    trisdecagon
)
from prysm.objects import (
    Slit,
    Pinhole,
    SiemensStar,
    TiltedSquare,
    SlantedEdge,
)

from prysm.lens import Lens

__all__ = [
    'config',
    'plot_fourier_chain',
    'Detector',
    'OLPF',
    'PixelAperture',
    'Pupil',
    'FringeZernike',
    'StandardZernike',
    'Interferogram',
    'Seidel',
    'PSF',
    'AiryDisk',
    'MTF',
    'Lens',
    'gaussian',
    'rotated_ellipse',
    'regular_polygon',
    'square',
    'pentagon',
    'hexagon',
    'heptagon',
    'octagon',
    'nonagon',
    'decagon',
    'hendecagon',
    'dodecagon',
    'trisdecagon',
    'Slit',
    'Pinhole',
    'SiemensStar',
    'TiltedSquare',
    'SlantedEdge',
    'Convolvable',
    'circle',
    'truecircle',
]



"""Extra functions, used for demos and education."""
from .util import share_fig_ax


def plot_fourier_chain(pupil, psf, mtf, fig=None, axs=None, sizex=12, sizey=6):
    '''Plot a `Pupil` `PSF`, and `MTF` demonstrating the process of fourier optics simulation.

    Parameters
    ----------
    pupil : `Pupil`
        The pupil of an optical system
    psf : `PSF`
        The psf of an optical system
    mtf : `MTF`
        The MTF of an optical system
    fig : `matplotlib.figure.Figure`, optional:
        A figure object
    axs : `list` of `matplotlib.axes.Axis`
        axes to place the plots in
    sizex : `float`
        size of the figure in x
    sizey : `float`
        size of the figure in y

    Returns
    -------
    `matplotlib.figure.Figure`
        A figure containing the plot
    `matplotlib.axes.Axis`
        An axis containing the plot

    '''
    fig, axs = share_fig_ax(fig, axs, numax=3)

    pupil.interferogram(fig=fig, ax=axs[0])
    psf.plot2d(fig=fig, ax=axs[1])
    mtf.plot2d(fig=fig, ax=axs[2])

    axs[0].set(title='Pupil')
    axs[1].set(title='PSF')
    axs[2].set(title='MTF')

    bbox_props = dict(boxstyle="rarrow", fill=None, lw=1)
    axs[0].text(1.385, 1.07, r'|Fourier Transform|$^2$',
                ha='center', va='center', bbox=bbox_props,
                transform=axs[0].transAxes)
    axs[0].text(3.15, 1.07, r'|Fourier Transform|',
                ha='center', va='center', bbox=bbox_props,
                transform=axs[0].transAxes)
    fig.set_size_inches(sizex, sizey)
    fig.tight_layout()
    return fig, axs
